#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END
"""
Tabular formatting of output
"""

import sys
import ujson as json

def _fmt_txt(columns, rng):
    fmts = ['{}' for x in columns[0]]
    maxs = [1 for x in columns[0]]
    for row in columns:
        for rnx in rng:
            if not isinstance(row[rnx], str):
                row[rnx] = str(row[rnx])
            len_cell = len(row[rnx])
            if len_cell > maxs[rnx]:
                maxs[rnx] = len_cell

    for rnx in rng:
        fmt = "{:" + str(maxs[rnx]) + "} "
        fmts[rnx] = fmt
    fmts[-1] = '{}\n'
    return fmts

def _fmt_tsv(columns, rng):
    fmts = ['{}' for x in columns[0]]

    for row in columns:
        for rnx in rng:
            val = row[rnx]
            if not isinstance(val, str):
                val = str(val)
            if "\t" in val:
                val = val.replace("\t", '\\t')
            row[rnx] = val

    for rnx in rng:
        fmts[rnx] = "{}\t"
    fmts[-1] = "{}\n"

    return fmts

def _fmt_csv(columns, rng):
    for row in columns:
        for rnx in rng:
            val = row[rnx]
            if not isinstance(val, str):
                val = str(val)
            if '"' in val:
                val = val.replace('"', '""')
            row[rnx] = val

    fmts = ['{}' for x in columns[0]]
    for rnx in rng:
        fmts[rnx] = "\"{}\","
    fmts[-1] = "\"{}\"\n"

    return fmts

###########################################################################
# pylint: disable=dangerous-default-value, too-many-locals
def cols(columns, header=False, stderr=list(), sort=None, fmt="txt"):
    """
    Print columns tabularly, supporting optional headers and mixed
    stdout/stderr streams.

    Format can be:

        txt - space separated (default)
        tsv - tab separated
        csv - comma separated

    >>> cols([[1,2,3], ["ba","bc","d"]])
    1  2  3
    ba bc d
    >>> def sort_second(row): return row[1]
    >>> cols([["longerthanlongerthanlong","ab","zz","d"], ["ba","bc","d", "z"]], sort=sort_second)
    longerthanlongerthanlong ab zz d
    ba                       bc d  z
    >>> cols([["dc","ab","zz","d"], ["ba","bc","d", "z"]], stderr=[1,2,3])
    ba
    dc
    >>> cols([["this","that","there","then"],
    ...       ["dc","ab","zz","d"],
    ...       ["ba","bc","d", "z"]], header=True)
    this that there then
    ba   bc   d     z
    dc   ab   zz    d
    """

    rng = range(0, len(columns[0]))

    if fmt == "list":
        fmt = "txt"

    if fmt == "json":
        sys.stdout.write(json.dumps(columns))
        return

    fmts = globals()["_fmt_" + fmt](columns, rng)

    def sort_first(rowx): # pylint: disable=missing-docstring
        return rowx[0]
    if not sort:
        sort = sort_first

    headers = list()
    if header:
        headers = columns[0]
        columns = columns[1:]

    def output_mixed(rowx, stderr): #pylint: disable=missing-docstring
        for rnx in rng:
            stream = sys.stdout
            if rnx in stderr or headers and headers[rnx] in stderr:
                stream = sys.stderr
            stream.write(fmts[rnx].format(rowx[rnx]))
            stream.flush()
        sys.stdout.flush()

    def output_stdout(rowx, stderr): #pylint: disable=missing-docstring,unused-argument
        sys.stdout.write(''.join(fmts).format(*rowx))

    if stderr:
        output = output_mixed
    else:
        output = output_stdout

    if headers:
        output_mixed(headers, rng) # all header goes to stderr

    for row in sorted(columns, key=sort):
        output(row, stderr)



#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END

"""
Reflex Engine Client

Wrapper handling Reflex Apikey

Future: add SSL client certificates
"""

import urllib
import time
import base64
import requests
import nacl.utils
#import uuid
import jwt
import dictlib
import rfx
from rfx import threadlock, json2data, json4human, json4store

################################################################################
class Unauthorized(Exception):
    """Client Error"""
    pass

################################################################################
class ClientError(Exception):
    """Client Error"""
    pass

################################################################################
# pylint: disable=too-many-instance-attributes
class Session(rfx.Base):
    """Session object, which logs into Reflex Engine and manages the session"""
    session_jti = None
    session_sid = None
    session_secret = None
    session_expires = 0
    apikey_name = None
    apikey_secret = None
    _cache = None
    headers = None

    def __init__(self, **kwargs):
        super(Session, self).__init__(**kwargs)
        base = kwargs.get('base')
        if base:
            rfx.Base.__inherit__(self, base)
        self._cache = dict()
        self.headers = dict()

    ############################################################################
    def _login(self, force=False):
        """Use the Apikey to get a session key"""
        if not force and self.session_expires > time.time():
            return

        if not self.apikey_name:
            name, b64_secret = (self.cfg['REFLEX_APIKEY'] + ".").split(".")[0:2]
            self.apikey_name = name
            self.apikey_secret = base64.b64decode(b64_secret)

        key_jwt = jwt.encode(dict(
            jti=self.apikey_name,
            #seed=str(uuid.uuid4()), #base64.b64encode(nacl.utils.random(256)).decode(),
            seed=base64.b64encode(nacl.utils.random(256)).decode(),
            exp=time.time() + 300
        ), self.apikey_secret)

        headers = self.headers.copy()
        headers["X-Apikey"] = key_jwt
        headers["Content-Type"] = "application/json"
        result = requests.get(self.cfg['REFLEX_URL'] + "/token", headers=headers)

        if result.status_code == 200:
            self.DEBUG("Authorized")
            data = json2data(result.content.decode())
            self.session_sid = data['session']
            self.session_jti = data['jti']
            self.session_secret = base64.b64decode(data['secret'])
            self.session_expires = data['expires_at']

        else:
            if self.do_DEBUG():
                self.DEBUG("Failed to authorize:\n\tHTTP {}\n{}\n\n{}\n".format(
                    result.status_code,
                    json4human(dict(result.headers)),
                    json2data(result.content.decode())))

            raise Unauthorized("Unable to authorize session")

    ############################################################################
    # pylint: disable=too-many-branches
    def _call(self, func, target, *args, **kwargs):
        """Call Reflex Engine, wrapped with authentication and session management"""
        try:
            self._login()
        except Unauthorized as err:
            self.ABORT("Unauthorized: " + str(err))
        except requests.exceptions.ConnectionError:
            self.ABORT("Unable to connect to REFLEX_URL ({})".format(self.cfg['REFLEX_URL']))

        # enrich the arguments
        headers = self.headers.copy()
        if kwargs.get('headers'):
            headers = dictlib.union(headers, kwargs['headers'])
        if not kwargs.get('cookies'):
            kwargs['cookies'] = {}
        if not headers.get('Content-Type'):
            headers['Content-Type'] = "application/json"
        kwargs['headers'] = headers

        query = self.cfg['REFLEX_URL'] + "/" + target
        if self.debug.get('remote-abac'):
            if "?" in query:
                query += "&"
            else:
                query += "?"
            query += "abac=log"

        # make the call
        result = self._call_sub(func, query, *args, **kwargs)

        # unlikely, as self._login() should take care of this, unless our timing
        # is off from the server's, but just in case...
        if result.status_code == 401:
            self.DEBUG("Unauthorized received, Retrying Login")
            self._login(force=True)
            result = self._call_sub(func, query, *args, **kwargs)

        if result.status_code == 500:
            raise ClientError("Server side error")

        if result.status_code == 404:
            raise ClientError("Endpoint or object not found (" + query + ")")

        if "application/json" not in result.headers.get('Content-Type', ''):
            self.DEBUG("error", result.content.decode())
            raise ClientError("Result is not valid content type")

        if result.status_code == 204:
            return {}

        if result.status_code in (200, 201, 202):
            return result.json()

        raise ClientError(result.json()['message'])

    ############################################################################
    def _call_sub(self, func, *args, **kwargs):
        """Subcall of call"""
        auth_jwt = jwt.encode(dict(
            jti=self.session_jti,
            exp=time.time() + 60
        ), self.session_secret)
        dbg = self.do_DEBUG()

        kwargs['cookies']['sid'] = self.session_sid
        if self.headers:
            kwargs['headers'].update(self.headers)
        kwargs['headers']['X-ApiToken'] = auth_jwt

        if dbg:
            self.DEBUG("auth", jwt=auth_jwt)
            self.DEBUG("call", func=func, args=args, kwargs=kwargs)

        return func(*args, **kwargs)

    ############################################################################
    def get(self, obj_type, obj_target, archive=False):
        """session GET"""
        args = []
        if archive:
            args.append("archive=" + str(archive['start']))
        return self._call(requests.get,
                          obj_type + "/" + str(obj_target) + "?" + "&".join(args))

    ############################################################################
    # pylint: disable=too-many-arguments
    def list(self, obj_type, match=None, cols=None, raise_error=True, archive=False):
        """
        session LIST.  Match is a glob pattern (optional), cols is a list
        of column names
        """
        args = []
        if match:
            try: # stupid python2
                args.append("match=" + urllib.parse.quote(match))
            except: # pylint: disable=bare-except, no-member
                args.append("match=" + urllib.pathname2url(match))
        if archive:
            args.append("archive=" +
                        str(archive['start']) + "~" +
                        str(archive['end']))
        if cols:
            args.append("cols=" + ",".join(cols))

        querystr = obj_type + "/"
        if args:
            querystr += "?" + "&".join(args)
        if raise_error:
            return self._call(requests.get, querystr)
        else:
            try:
                return self._call(requests.get, querystr)
            except: # pylint: disable=bare-except
                return list()

    ############################################################################
    def create(self, obj_type, obj_data):
        """session CREATE"""
        return self._call(requests.post, obj_type, data=json4store(obj_data))

    ############################################################################
    def update(self, obj_type, obj_target, obj_data):
        """session UPDATE"""
        path = obj_type + "/" + str(obj_target)
        return self._call(requests.put, path, data=json4store(obj_data))

    ############################################################################
    def patch(self, obj_type, obj_target, obj_data):
        """session PATCH"""
        path = obj_type + "/" + str(obj_target) + "?merge=true"
        return self._call(requests.put, path, data=json4store(obj_data))

    ############################################################################
    def instance_ping(self, obj_target, obj_data):
        """special instance update ping"""
        path = "instance-ping/" + str(obj_target)
        return self._call(requests.put, path, data=json4store(obj_data))

    ############################################################################
    def delete(self, obj_type, obj_target):
        """session DELETE"""
        return self._call(requests.delete, obj_type + "/" + str(obj_target))

    ############################################################################
    @threadlock
    def cache_get(self, obj_type, obj_target, **kwargs):
        """
        Cache wrapper around .get()
        """
        if obj_type in self._cache:
            if obj_target in self._cache[obj_type]:
                return self._cache[obj_type][obj_target]
        else:
            self._cache[obj_type] = dict()
        try:
            obj = self.get(obj_type, obj_target, **kwargs)
        except ClientError:
            self._cache[obj_type][obj_target] = None
            raise
        self._cache[obj_type][obj_target] = obj
        return obj

    ############################################################################
    @threadlock
    def cache_update(self, obj_type, obj_target, payload, **kwargs):
        """
        Cache wrapper around .update()
        """
        value = self.update(obj_type, obj_target, payload, **kwargs)
        if value:
            if obj_type in self._cache:
                if obj_target in self._cache[obj_type]:
                    del self._cache[obj_type][obj_target]
        return value

    ###########################################################################
    @threadlock
    def cache_list(self, obj_type, **kwargs):
        """
        Cache wrapper around .list()
        """
        if not obj_type in self._cache:
            self._cache[obj_type] = {}

        objs = self.list(obj_type, **kwargs)
        for obj in objs:
            oname = obj.get('name')
            if oname:
                self._cache[obj_type][oname] = obj

        return objs

    ###########################################################################
    @threadlock
    def cache_reset(self):
        """Clear the cache"""
        self._cache = dict()

    ###########################################################################
    def cache_drop(self, obj_type, obj_target):
        """Drop an object from the cache, if it exists"""
        if obj_type in self._cache and obj_target in self._cache[obj_type]:
            del self._cache[obj_type][obj_target]



#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Service Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END

"""
General reflex module and controls
"""

import io
import os
import os.path
import pprint # used by debug
import sys
import time
import re
import copy
import fcntl # get_my_ips, unbuffered
import struct # get_my_ips
import socket # get_my_ips
import threading
import traceback
import subprocess
import base64
from stat import S_IRWXU # LOG2FILE
import requests
import nacl.utils
import jwt
import ujson as json
# cipher bits
from rfx import crypto

# find the reflex base install folder based on this lib
BASEDIR = os.path.realpath(os.path.dirname(__file__)).rsplit("/", 1)[0]
while BASEDIR and not os.path.exists(BASEDIR + "/.pkg/version"):
    BASEDIR = BASEDIR.rsplit("/", 1)[0]

################################################################
def json4human(data):
    """Json output for humans"""
    return json.dumps(data, escape_forward_slashes=False, indent=2, sort_keys=True)

def json4store(data, **kwargs):
    """Json output for storage"""
    return json.dumps(data, escape_forward_slashes=False, **kwargs)

def json2data(string):
    """json to its python representation"""
    return json.loads(string)

################################################################
def threadlock(func):
    """use a thread lock on current method, if self.lock is defined"""
    def threadlock_wrapper(*args, **kwargs):
        """Decorator Wrapper"""
        lock = args[0].lock
        lock and lock.acquire(True) # pylint: disable=expression-not-assigned
        try:
            result = func(*args, **kwargs)
            lock and lock.release() # pylint: disable=expression-not-assigned
        except:
            lock and lock.release() # pylint: disable=expression-not-assigned
            raise

        return result

    return threadlock_wrapper

################################################################
def get_my_ips():
    """highly os specific - works only in modern linux kernels"""
    ips = list()
    for ifdev in os.listdir("/sys/class/net"):
        if ifdev == "lo":
            continue
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            ips.append(socket.inet_ntoa(fcntl.ioctl(
                sock.fileno(),
                0x8915,  # SIOCGIFADDR
                struct.pack('256s', ifdev[:15].encode())
            )[20:24]))
        except OSError:
            pass
    return ips

################################################################
# pylint: disable=too-few-public-methods
class Colorize(object):
    """Enhance an object with ANSI colorizing"""

    colorize = False
    ansi_codes = {
        'start': "\033[",
        'fgblue': '34m',
        'fgred': '31m',
        'fggreen': '32m',
        'fggray': '30m',
        'reset': '0m',
    }

    ############################################################
    def ansi_code(self, code):
        """
        If colorizing is enabled, return the requested ansi code
        """
        if not self.colorize:
            return ''
        return self.ansi_codes['start'] + self.ansi_codes[code]

################################################################
# pylint: disable=too-many-instance-attributes, too-many-public-methods
class Base(Colorize):
    """Base object where common and global attributes are defined"""

    cfg = {
        "REFLEX_URL": "",
        "REFLEX_APIKEY": ""
    }
    cmd = 'reflex'
    logfmt = 'txt'
    logfd = sys.stderr
    outputfd = sys.stdout
    notifyfd = sys.stderr
    debug = {'basic':False}
    timestamp = False
    term_width = 80
    lock = None # used with threadsafe=True init arg
    my_ip = ''
    my_ips = None
    my_host = ''

    # global static. DO NOT CHANGE. This should be enhanced with a local secret file
    secret_seed = "Y+8CcVDNH/HWt1nDuNPrdl0npPQYwrPZ3ZqSMtutbso="

    ############################################################
    def __inherit__(self, base):
        """To avoid multiple config loads, call __inherit__ to inherit a parent's previous load"""
        self.logfmt = base.logfmt
        self.DEBUG("rfx.Base.__inherit__(base={0})".format(base))
        self.debug = base.debug
        self.cfg = base.cfg
        self.timestamp = base.timestamp
        self.logfd = base.logfd
        self.outputfd = base.outputfd
        self.notifyfd = base.notifyfd
        # do not copy self.lock

        return self

    # pylint: disable=unused-argument
    def __init__(self, *args, **kwargs):
        self.logfmt = kwargs.get('logfmt', 'txt')
        os.environ['LOGFMT'] = self.logfmt
        #self.DEBUG("rfx.Base.__init__()")
        if 'debug' in kwargs and kwargs['debug']:
            for module in kwargs['debug']:
                self.debug[module] = True

        # implement a mutex if requested
        if kwargs.get('threadsafe'):
            self.lock = threading.Lock()

    ############################################################
    def get_my_nameip(self):
        """Try to get my own IP, capture errors"""
        self.my_host = socket.gethostname()
        try:
            self.my_ips = get_my_ips()
            self.my_ip = self.my_ips[0]
        except: # pylint: disable=bare-except
            try:
                self.my_ip = socket.gethostbyname(self.my_host)
                self.my_ips = [self.my_ip]
            except: # pylint: disable=bare-except
                self.NOTIFY("Not able to identify my own address")
                self.my_ip = ''
                self.my_ips = []

    ############################################################
    # terminal size
    # pylint: disable=no-self-use
    def get_terminal_width(self):
        """return the terminal height, width"""
        sub = subprocess.Popen(["stty", "size"],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
        sub.wait()
        if sub.returncode > 0:
            return
        try:
            width = sub.stdout.read().split()[1]
            self.term_width = int(width) or 80
        except IndexError: # pylint: disable=broad-except
            self.NOTIFY("Unable to calculate terminal width")

    ############################################################
    def cfg_load(self):
        """Load state config data"""
        try:
            self.DEBUG("Config Load:", data=self.cfg, module='config')
            cfg_path = os.environ['HOME'] + '/.reflex/cfg'
            if os.path.exists(cfg_path):
                with open(cfg_path, 'rb') as in_file:
                    data = in_file.read().decode()

                #vers = data[:4]
                data = data[4:]

                # do something with version here
                key = crypto.Key(self.cfg_secret())
                cipher = crypto.Cipher(key)
                self.cfg = json.loads(cipher.key_decrypt(data).encode())

            else:
                self.DEBUG("No .reflex/cfg", module='config')

            # then pull environment settings, as they have higher priority
            self._env2cfg('REFLEX_URL')
            self._env2cfg('REFLEX_APIKEY')
        except ValueError as err:
            self.NOTIFY("Unable to load config!")
        except IOError as err:
            self.NOTIFY("Unable to load config!")
            self.DEBUG("Reason: " + str(err))

        if not self.cfg.get('REFLEX_URL') or not self.cfg.get('REFLEX_APIKEY'):
            self.NOTIFY("neither REFLEX_URL nor REFLEX_APIKEY configured, try: reflex setup wizard")
        return self

    ############################################################
    def cfg_save(self):
        """Save config to file"""
        base = os.environ['HOME'] + '/.reflex'
        if not os.path.exists(base):
            os.mkdir(base)
            os.chmod(base, 0o0700)
        elif not os.path.isdir(base):
            self.ABORT("Config folder {} is not a directory!"
                       .format(base))

        cfg_file = base + '/cfg'
        data = json4store(self.cfg)
        try:
            key = crypto.Key(self.cfg_secret())
            with open(cfg_file, 'wb') as out_file:
                os.chmod(cfg_file, 0o0600)
                # future safe: so we can handle different versions and keys
                out_file.write("0000".encode())
                cipher = crypto.Cipher(key)
                out_file.write(cipher.key_encrypt(data).encode())

        except IOError as err:
            self.NOTIFY("Unable to save config!")
            self.DEBUG("Reason: " + str(err))

    ############################################################
    # security through obscurity
    def cfg_secret(self):
        """Pull the secret key.  Not great security, but better than non-encrypted"""
        site_key_file = os.environ.get('REFLEX_SITE_KEY')
        if site_key_file and os.path.exists(site_key_file):
            with open(site_key_file, 'r') as in_file:
                self.DEBUG("Using site crypto key", module='crypto')
                return base64.b64decode(in_file.read())

        return self.secret_seed

    ############################################################
    def cfg_get(self, name, default=None, required=False):
        """
        Pull a CFG value, but allow the environment to override
        >>> b = Base()
        >>> b.cfg['TST'] = 'sue'
        >>> b.cfg_get('TST')
        'sue'
        >>> os.environ['TST'] = 'bob'
        >>> b.cfg_get('TST')
        'bob'
        >>> b.cfg_get('ZZZZZZ', default='no')
        'no'
        >>> b.cfg_get('ZZZZZZ', required=True)
        Traceback (most recent call last):
        ...
        ValueError: ZZZZZZ not defined (in environ or reflex settings)
        """
        if name in os.environ.keys():
            return os.environ[name]
        elif name in self.cfg.keys():
            return self.cfg[name]
        elif default:
            return default
        elif required:
            raise ValueError(name + " not defined (in environ or reflex settings)")
        else:
            return ''

    ############################################################
    def _env2cfg(self, key):
        """Replace config key from environment or docker secrets, if it exists"""
        secret_path = "/run/secrets/" + key
        if os.path.exists(secret_path):
            with open(secret_path) as infile:
                self.cfg[key] = infile.read().strip()
        elif key in os.environ.keys():
            self.cfg[key] = os.environ[key]

    ############################################################
    # pylint: disable=invalid-name,too-many-branches
    def NOTIFY(self, *msg, **kwargs):
        """
        Internal print wrapper, so it can be easily overloaded--Notify is for human readable

        >>> b = Base()
        >>> b.notifyfd = sys.stdout
        >>> b.NOTIFY("Hello")
        Hello
        """
        if len(msg) > 1:
            msg = " ".join(msg)
        elif msg:
            msg = msg[0]
        else:
            msg = ''

        if self.logfmt == 'json':
            merged = self.union_dict({
                'timestamp':int(time.time()*1000),
                'service':self.__class__.__name__,
            }, kwargs)
            if 'msg' in merged and not merged['msg']:
                del merged['msg']
            if msg:
                if not merged.get('msg'):
                    merged['msg'] = msg.join(" ")
                else:
                    merged['_msg'] = msg.join(" ")
            json.dump(merged, self.notifyfd, escape_forward_slashes=False)
            return

        msg_start = ''
        msg_end = '\n'
        if 'color' in kwargs:
            msg_start = self.ansi_code('fg' + kwargs['color'])
            msg_end = self.ansi_code('reset') + '\n'
            del kwargs['color']

        if self.timestamp:
            msg_start += self.TIMESTAMP()

        if kwargs:
            for kwarg in kwargs:
                val = str(kwargs[kwarg])
                if " " in val:
        # see engine/logging.txt message
        #            val = json4store(kwargs[kwarg])
                    val = '"' + re.sub(r'\n', '\n\t', val.replace('"', '\\"')) + '"'
                msg += " " + kwarg + "=" + val

        self.notifyfd.write(msg_start + msg + msg_end)

    ############################################################
    def TIMESTAMP(self): # pylint: disable=invalid-name
        """return a timestamp"""
        return time.strftime("%Y-%m-%d %H:%M:%S ")


    ############################################################
    def OUTPUT(self, msg): # pylint: disable=invalid-name
        """
        Internal print wrapper, so it can be easily overloaded--OUTPUT is for machine readable

        >>> b = Base()
        >>> b.outputfd = sys.stdout
        >>> b.OUTPUT("Hello")
        Hello
        """
        if isinstance(msg, bytes): # grr python 2/3
            msg = msg.decode()
        self.outputfd.write(msg + "\n")

    ############################################################
#    def _json_logmsg(self, **args):
#        merged = self.union_dict({
#            'timestamp':int(time.time()*1000),
#            'service':self.__class__.__name__,
#        }, args)
#        if not len(merged['message']):
#            del merged['message']
#        return json4store(merged)

    ############################################################
    ######## Full request debugging, put this in the appropriate places:
    #        import logging
    #        import httplib
    #        httplib.HTTPConnection.debuglevel = 1
    #        # you need to initialize logging, otherwise you will
    #        # not see anything from requests
    #        logging.basicConfig()
    #        logging.getLogger().setLevel(logging.DEBUG)
    #        requests_log = logging.getLogger("requests")
    #        requests_log.setLevel(logging.DEBUG)
    #        requests_log.propagate = True
    ########

    def do_DEBUG(self, module=""): # pylint: disable=invalid-name
        """Am I in a debug mode? """
        if not module:
            module = str(self.__class__.__name__).lower()
        debug = self.debug or {'*':False}
        if debug.get('*', None) or module in debug.keys():
            return True
        return False

    # pylint: disable=invalid-name,too-many-arguments
    def DEBUG(self, msg, module="", err=None, color='gray', **kwargs):
        """Send debug output, if we are in debug mode"""
        if not module:
            module = str(self.__class__.__name__)
        if self.do_DEBUG(module=module):
            if err:
                msg = msg + ": " + str(err)
            self.NOTIFY("DEBUG " + module + ": " + msg,
                        level="DEBUG", err=err, color=color, **kwargs)

    ############################################################
    def ABORT(self, msg, err=None): # pylint: disable=invalid-name
        """Abort and fail nicely"""
        if err:
            self.DEBUG("Error", err=err, color='red')
        self.NOTIFY("ABORT: " + msg, color='red')
        if err and err.strerror:
            self.NOTIFY(str(err), color='red')

        sys.exit(1)

    ############################################################
    def LOG(self, **args): # pylint: disable=invalid-name
        """
        >>> b = Base()
        >>> b.logfd = sys.stdout
        >>> b.logfmt = 'txt'
        >>> args = {'x':"this is a test", 'y':'this', 'z':{"hello":1}}
        >>> b.LOG(**args) # doctest: +ELLIPSIS
        ts=... x="this is a test" y=this z={"hello":1}
        >>> b.logfmt = 'json'
        >>> b.LOG(**args) # doctest: +ELLIPSIS
        {".ts":...,"x":"this is a test","y":"this","z":{"hello":1}}
        """
        if self.logfmt == 'txt':
            buf = "ts=" + str(int(time.time()*1000))
            for key in sorted(args):
                value = args[key]
                if isinstance(value, str):
                    if ' ' in value:
                        if '"' in value:
                            value = value.replace('"', r'\"')
                        buf += " " + str(key) + '="' + value + '"'
                    else:
                        buf += " " + str(key) + "=" + value
                else:
                    # note: this will not parse well in splunk
                    buf += " " + str(key) + '=' + json4store(value)
        else:
            args['.ts'] = int(time.time()*1000)
            buf = json4store(args, sort_keys=True)
        self.logfd.write(buf + "\n")

    ############################################################
    def LOG2FILE(self, fname, mode, data): # pylint: disable=invalid-name
        """Mode should be one of: w, w+"""
        tmpdir = self.cfg_get('TMPDIR', default='/tmp')
        path_fname = tmpdir + "/" + fname
        with open(path_fname, mode) as outf:
            outf.write(json4human(data))
        if os.path.exists(path_fname):
            os.chmod(path_fname, S_IRWXU)

    ############################################################
    def union_dict(self, dict1, dict2):
        """
        Helper function - deep merge of two dictionaries

        >>> b = Base()
        >>> d = b.union_dict({"a":{"b":{"c":1}}},
        ...                  {"a":{"b":{"d":2}},"e":3})
        >>> import json
        >>> json.dumps(d, sort_keys=True)
        '{"a": {"b": {"c": 1, "d": 2}}, "e": 3}'
        """
        for key in dict1: # , value in dict1.iteritems():
            #key = str(key) # grr unicode
            value = dict1[key]
            if key not in dict2:
                dict2[key] = value
            elif isinstance(value, dict):
                self.union_dict(value, dict2[key])
        return dict2

    ############################################################
    def union_set(self, set1, set2): # pylint: disable=no-self-use
        """
        Helper function - union of two sets

        Defined as a method only so it can be called programatically

        """
        return set(set2).union(set(set1))

    ############################################################################
    def sed_env_dict(self, dictionary):
        """
        reference dictionary for sed_env
        """
        for key in dictionary:
            dictionary[key] = self.sed_env(dictionary[key], dictionary, key)
        return dictionary

    ############################################################################
    # pylint: disable=dangerous-default-value
    def sed_env(self, value, dictionary, source_value, env={}):
        """
        search and replace ${VARIABLE} in stream, using os.environ for vars,
        or keyword argument 'env' if it is defined
        """
        if not env:
            env = os.environ

        def env_match(match):
            """sub function used in matching"""
            match_key = match.group(1)
            if match_key in env:
                return env[match_key]
            elif match_key != source_value and match_key in dictionary:
                return dictionary[match_key]
            return None
        return re.sub(r"\$\{([a-zA-Z0-9_-]+)\}", env_match, value)

    ###########################################################################
    def TRAP(self, func, *args, **kwargs):
        """
        Call a function and ignore any thrown errors.  See errors with --debug
        """
        try:
            self.DEBUG("(| {}({}, {}) |)".format(func, args, kwargs))
            return func(*args, **kwargs)
        except: # pylint: disable=bare-except
            self.DEBUG(traceback.format_exc())
            return False

###############################################################################
# Common Errors we might throw
# pylint: disable=missing-docstring
class NotFoundError(Exception):
    pass

# pylint: disable=missing-docstring
class CannotContinueError(Exception):
    pass

def unbuffer(filed):
    fcl = fcntl.fcntl(filed.fileno(), fcntl.F_GETFL)
    fcl |= os.O_SYNC
    fcntl.fcntl(filed.fileno(), fcntl.F_SETFL, fcl)

################################################################################
def set_interval(milliseconds, func, *args, **kwargs):
    """
    Call function every interval.  Starts the timer at call time.
    Although this could also be a decorator, that would not initiate the time at
    the same time, so would require additional work.

    Arguments following function will be sent to function.  Note that these args
    are part of the defining state, and unless it is an object will reset each
    interval.

    The inine test will print "TickTock x.." every second, where x increments.

    >>> import time
    >>> class Tock(object):
    ...     count = 0
    ...     stop = None
    >>> def tick(obj):
    ...     obj.count += 1
    ...     if obj.stop and obj.count == 4:
    ...         obj.stop.set() # shut itself off
    ...         return
    ...     print("TickTock {}..".format(obj.count))
    >>> tock = Tock()
    >>> tock.stop = set_interval(1000, tick, tock)
    >>> time.sleep(6)
    TickTock 1..
    TickTock 2..
    TickTock 3..
    """

    stopper = threading.Event()
    def interval(seconds, func, *args, **kwargs):
        def wrapper():
            if stopper.isSet():
                return
            interval(seconds, func, *args, **kwargs)
            try:
                func(*args, **kwargs)
            except: # pylint: disable=bare-except
                traceback.print_exc()

        thread = threading.Timer(seconds, wrapper)
        thread.daemon = True
        thread.start()
    interval(milliseconds/1000, func, *args, **kwargs)
    return stopper



#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END

"""
Powerfully easy argument parsing
"""

import sys
import copy
# subprocess and input are used by password
import subprocess
# pylint: disable=duplicate-code
try:
    from builtins import input # pylint: disable=redefined-builtin
    get_input = input # pylint: disable=invalid-name
except: # pylint: disable=bare-except
    get_input = raw_input # pylint: disable=invalid-name, undefined-variable

################################################################################
class ArgParse(Exception):
    """Error when processing args"""
    pass


############################################################################
def opt_key(optname, optinfo):
    """
    Return the key name to use for an option

    >>> opt_key("bo?bo", {'key': 'nono'})
    'nono'
    >>> opt_key("bo?bo", {})
    'bobo'
    """
    if optinfo.get('key'):
        return optinfo.get('key')
    return strip_opt(optname)

################################################################################
def arg_key(argname, arginfo):
    """
    Return the key name to use for an argument

    >>> arg_key("bobo", {'key': 'nono'})
    'nono'
    >>> arg_key("bobo", {})
    'bobo'
    """
    if arginfo.get('key'):
        return arginfo.get('key')
    return argname

################################################################################
def from_set(name, info, value):
    """
    Select a value from a set, using option matching

    >>> from_set("thegang", {'set': ["scoob?y", "fred", "vel?ma"]}, "shaggy")
    Traceback (most recent call last):
    ...
    optarg.ArgParse: Invalid option for thegang (shaggy), must be one of: scoob?y, fred, vel?ma
    >>> from_set("thegang", {'set': ["scoob?y", "fred", "vel?ma"]}, "velm")
    'velma'
    """
    vals = info.get('set')
    if not vals:
        raise ArgParse("No set defined for arg type=from-set")
    if isinstance(vals, list):
        for opt in vals:
            opt = str(opt)
            if opt == "*":
                return value
            if optmatch(opt, value):
                return strip_opt(opt)
        raise ArgParse("Invalid option for " + name + " (" + value +
                       "), must be one of: " + ", ".join(vals))
    else:
        raise ArgParse("Invalid definition, 'set' value is not list") # or Args")

################################################################################
def strip_opt(name):
    """
    strip special characters from an option definition

    >>> strip_opt("fr?ed|sc?ooby|shaggy")
    'fred'
    """
    return name.split("=", 1)[0].split("|", 1)[0].replace("?", "")

################################################################################
def optmatch(pattern, match):
    """
    Easier argument specs than argparse

    >>> optmatch("l?ist|ls", "lis")
    True
    >>> optmatch("l?ist|ls", "ls")
    True
    >>> optmatch("l?ist|ls", "l")
    True
    >>> optmatch("l?ist|ls", "loo")
    False
    """
    match = str(match).lower().split("=", 1)[0]
    for variant in pattern.split("|"):
        req, opt = (variant + "?").split("?")[0:2]
        if req == match:
            return True
        for char in opt:
            req = req + char
            if req == match:
                return True
    return False

################################################################################
class Args(object):
    # pylint: disable=line-too-long
    """
    Define an argument set.  Provide a list of argument definitions.
    Arg definitions are a two array element:

        [ "name", {info} ]

    Order is important, as it relates to the order of sys.argv.  Arguments
    which are not processed are preserved, and may be later passed to a
    secondary args object for futher processing (such as with forked command
    sets).

    {info} has "type" defined, as one of:
        "from-set", "set-value", "set-add", "set-true", "set-password"
    if type is from-set, the additional key "set" is defined as an array,
    with strings being any match strings for optmatch()

    Result of calling args.handle_parse() is a dictionary of matched values.

    May call subsequent parsers and pass along already parsed information by
    passing along argv=args.argv and ops={result of previous parse call}

    >>> import json
    >>> def dump(data): return json.dumps(data, sort_keys=True)
    >>> args = Args(
    ...     [
    ...         "action", {
    ...             "type": "from-set",
    ...             "set": ["d?rive|a?ccelerate", "stop|b?rake"]
    ...         }
    ...     ], [
    ...         "where", {
    ...             "type": "set-value"
    ...         }
    ...     ], [
    ...         "--over?ride|-o", {
    ...             "type": "store-true"
    ...         }
    ...     ]
    ... )
    >>> dump(args.args)
    '[["action", {"set": ["d?rive|a?ccelerate", "stop|b?rake"], "type": "from-set"}], ["where", {"type": "set-value"}]]'
    >>> dump(args.opts)
    '{"--h?elp|-h": {"type": "syntax"}, "--over?ride|-o": {"type": "store-true"}}'
    >>> dump(args.handle_parse(argv=["drive", "north", "south", "--overr", "-p"]))
    '{"--override": true, "action": "drive", "where": "north"}'
    >>> args.argv
    ['south', '--overr', '-p']
    """
    args = None
    opts = None
    argv = None
    out = None

    ############################################################################
    def __init__(self, *arginput):
        self.args = list()
        self.opts = dict()
        self.opts["--h?elp|-h"] = {"type":"syntax"}

        for opt, dat in arginput:
            if opt[0] == '-':
                self.opts[opt] = dat
            else:
                self.args.append([opt, dat])

    ############################################################################
    def handle_parse(self, caller=None, argv=None, out=None, opts=None):
        """
        Top level method called to start parsing.
        Traps the ArgParse exception with a nice print.
        """
        try:
            return self.parse(caller, argv=argv, out=out, opts=opts)
        except ArgParse as err:
            print(err)
            sys.exit(1)

    ############################################################################
    # pylint: disable=too-many-branches
    def parse(self, caller=None, argv=None, out=None, opts=None):
        """
        Lower level parser, may be called at multiple levels
        """
        if isinstance(out, dict):
            self.out = out
        else:
            self.out = dict()


        # inherit options from higher levels
        if isinstance(opts, dict):
            for opt, val in opts.items():
                if opt[0] == '-':
                    self.out[opt] = val

        if isinstance(argv, list):
            self.argv = copy.copy(argv)
        else:
            self.argv = copy.copy(sys.argv[1:])

        unparsed = list()
        while self.argv:
            arg = self.argv.pop(0)

            if arg[0] == '-':
                matched = False
                for opt in self.opts:
                    if optmatch(opt, arg):
                        matched = True
                        self._set_opt(caller, opt, arg)
                        break
                if not matched:
                    unparsed.append(arg)
            else:
                if self.args:
                    self._set_arg(arg)
                else:
                    unparsed.append(arg)
                    continue

        if self.args:
            if caller:
                caller.fail("Missing argument: {" + self.args[0][0] + "}")
            else:
                sys.exit("Missing argument: {" + self.args[0][0] + "}")

        self.argv = unparsed
        return self.out

    ############################################################################
    def _set_opt(self, caller, opt, arg):
        """
        Set an option
        """
        optinfo = self.opts[opt]
        opttype = optinfo.get('type', 'set-true')
        optkey = opt_key(opt, optinfo)
        if opttype in ('set-value', 'set-add', 'from-set'):
            if '=' in arg:
                arg, value = arg.split("=", 1)
            elif self.argv:
                value = self.argv.pop(0)
            else:
                raise ArgParse("Unable to get value for " + opt)
            opttype = optinfo.get('type', 'set-true')
            if opttype == 'from-set':
                value = from_set(opt, optinfo, value)
            elif opttype == 'set-add':
                if self.out.get(optkey):
                    self.out[optkey].append(value)
                    value = self.out[optkey]
                else:
                    value = [value]

        elif opttype == 'set-password':
            if '=' in arg:
                raise ArgParse(opt + " may only be queried via stdin")
            # ugly hack, could do this better
            current = self.out.get(optkey) or []
            clen = len(current) + 1
            subprocess.call(["stty", "-echo"])
            try:
                sys.stderr.write("Password {}: ".format(clen))
                sys.stderr.flush()
                pwd = get_input()
                sys.stderr.write("\n")
            except:
                raise
            finally:
                subprocess.call(["stty", "echo"])

            if current:
                self.out[optkey].append(pwd)
                value = self.out[optkey]
            else:
                value = [pwd]

        elif opttype == 'syntax':
            if caller:
                caller.fail()
            else:
                print("Help not supported")

        else:
            value = True # pylint: disable=redefined-variable-type

        self.out[optkey] = value

    ############################################################################
    def _set_arg(self, value):
        """
        Set an argument
        """
        argname = self.args[0][0]
        arginfo = self.args[0][1]
        self.args = self.args[1:]

        argtype = arginfo.get('type', 'set-value')
        if argtype == 'from-set':
            value = from_set(argname, arginfo, value)

        self.out[arg_key(argname, arginfo)] = value




#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END

"""
General Test Tools
"""

#from __future__ import absolute_import, division, print_function

# py2 backwards compatability
#from builtins import bytes
import os
import re
import hashlib
from types import ModuleType
import sys
import time
import subprocess
import traceback
import doctest
import rfx

###############################################################################
def check_file(fname, expected):
    """
    Match checksum on file as expected.
    """
    with open(fname, 'rb') as file_h:
        cksum = md5(file_h.read())
        if cksum != expected:
            return False
    return True

###############################################################################
def md5(data):
    """
    Make a hash
    """
    return hashlib.md5(data).hexdigest()

###############################################################################
def check_data(data, expected):
    """
    Match checksum on file as expected.
    """
    cksum = md5(data)
    if cksum != expected:
        return False
    return True

###############################################################################
class TAP(rfx.Base):
    """
    Simple Test Anything Protocol (TAP) output
    """
    _tap_good = 0
    _tap_bad = 0
    _tap_total = 0
    _tap_logfile = 'test.log'

    ###########################################################################
    # pylint: disable=super-init-not-called
    def __init__(self):
        os.environ['PYTHONPATH'] = ":".join(sys.path)
        self.colorize = True

    ###########################################################################
    # change behavior
    # pylint: disable=super-init-not-called
    def TRAP(self, func, *args, **kwargs):
        try:
            func(*args, **kwargs)
            return True
        except: # pylint: disable=bare-except
            self.log("--- Error Calling:\n    {}({}, {})\n--- "
                     .format(func, args, kwargs))
            self.log(traceback.format_exc())
            return False

    ###########################################################################
    # pylint: disable=invalid-name, no-else-return
    def ok(self, label, value):
        """
        Basic TAPS evaluation
        """
        self._tap_total += 1
        tstamp = "{:.7f} ".format(time.time())
        msg = "ok {} {}".format(self._tap_total, label)
        if value:
            self._tap_good += 1
            self.NOTIFY(tstamp + msg, color='green')
            self.log("==> " + msg + "\n\n")
            return True
        else:
            self._tap_bad += 1
            self.NOTIFY(tstamp + "not " + msg, color='red')
            self.log("==> not " + msg + "\n\n")
            return False

    ###########################################################################
    def ok_run(self, label, expected, exec_args, shell=False):
        """
        Run a command
        Failure if returncode != 0 or expected md5 does not match output
        """
        sub = subprocess.Popen(exec_args,
                               shell=shell,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
        sub.wait()
        output = sub.stdout.read()
        self.log_header(label)
        self.log_output("OUTPUT", output)
        if sub.returncode > 0:
            self._debug_output(output, force=True)
            self.OUTPUT(label + " return code failure ({})".format(sub.returncode))
            return self.ok(label, False)

        return self._ok_output_md5(label, expected, output)

    ###########################################################################
    def ok_run_compare(self, label, exc, *expected, shell=False):
        """
        Run a command
        Failure if returncode != 0 or expected md5 does not match output
        """
        sub = subprocess.Popen(exc,
                               shell=shell,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
        sub.wait()
        output = sub.stdout.read().decode()

        return self._ok_output_compare(label, output, *expected)

    ###########################################################################
    def log_output(self, label, output):
        """
        Log a block of text
        """
        def hdr(label): # pylint: disable=missing-docstring
            length = int((80 - (len(label))) / 2)
            self.log(("<" * length) + label + (">" * length) + "\n")
        self.log("\n")
        hdr(" BEGIN " + label + " ")
        self.log(output)
        hdr(" END " + label + " ")
        self.log("\n")

    ###########################################################################
    def _debug_output(self, output, force=False):
        """
        Print output if we are debugging it
        """
        if force or self.do_DEBUG(module='test'):
            if isinstance(output, bytes): # grr python 2/3
                output = output.decode()
            self.log_output("OUTPUT", output)

    ###########################################################################
    def ok_func(self, label, func, *args, **kwargs):
        """
        TAPS evaluation, calling function
        """
        return self.ok(label, self.TRAP(func, *args, **kwargs))

    ###########################################################################
    def ok_func_io(self, label, expected, base_class, func, *args, **kwargs):
        """
        TAPS evaluation, hashing result of io streams, assuming rfx.base object

        Assert: base_class.outputfd and notifyfd are already set to a StringIO
        """
        self.TRAP(func, *args, **kwargs)
        base_class.outputfd.seek(0)
        base_class.notifyfd.seek(0)
        output = base_class.outputfd.read() + base_class.notifyfd.read()
        return self._ok_output_md5(label, expected, output)

    ###########################################################################
    # pylint: disable=too-many-arguments
    def ok_func_compare(self, label, base_class, func, fargs, fkwargs, *expected, **kwargs):
        """
        TAPS evaluation, hashing result of io streams, assuming rfx.base object

        Assert: base_class.outputfd and notifyfd are already set to a StringIO
        """
        try:
            func(*fargs, **fkwargs)
        except: # pylint: disable=bare-except
            base_class.NOTIFY(traceback.format_exc())
        base_class.outputfd.seek(0)
        base_class.notifyfd.seek(0)
        output = base_class.outputfd.read() + base_class.notifyfd.read()
        base_class.outputfd.seek(0)
        base_class.notifyfd.seek(0)
        base_class.outputfd.truncate(0)
        base_class.notifyfd.truncate(0)
        return self._ok_output_compare(label, output, *expected, **kwargs)

    ###########################################################################
    def _ok_output_compare(self, label, output, *expected, **kwargs):
        """
        Compare object
        """
        def trim(string):
            """sub"""
            string = string.strip()
            string = re.sub(r'^\s+', '', string, flags=re.IGNORECASE|re.MULTILINE)
            string = re.sub(r'\s+$', '', string, flags=re.IGNORECASE|re.MULTILINE)
            return string

        negate = False
        if kwargs.get('negate'):
            negate = True
        self.log_header(label)
        self.log("================ Expect ================\n")
        if negate:
            self.log("<<NOT>> ")
        self.log("\n<<AND>>\n".join(expected) + "\n")
        self.log("=============== Received ===============\n")
        self.log(output)
        self.log("========================================")
        output = trim(output)
        if output[-1:] != "\n":
            self.log("\n")

        def rxsearch(rx):
            """sub"""
            rx = trim(rx)
            if re.search(rx, output):
                return not negate
            return negate

        truth = list(map(rxsearch, expected))
        # pylint: disable=redefined-variable-type
        truth = bool(len(set(truth)) == 1 and truth[0] is True)
        return self.ok(label, truth)

    ###########################################################################
    def _ok_output_md5(self, label, expected, output):
        """
        Check output to match md5, fail or not
        """
        cksum = md5(output)
        self._debug_output(output)
        self.log_header(label)
        self.log_output("OUTPUT", output)

        if cksum != expected:
            self.OUTPUT("Expected: {} != found: {}".format(expected, cksum))
            return self.ok(label, False)
        return self.ok(label, True)

    ###########################################################################
    def ok_func_cksum(self, label, expected, func, *args, **kwargs):
        """
        TAPS evaluation, hashing result of function call
        """
        result = self.TRAP(func, *args, **kwargs)
        self.DEBUG("=> {}".format(result))
        cksum = md5(str([result]))
        if cksum != expected:
            self.DEBUG("Expected: {} != found: {}".format(expected, cksum))
            return self.ok(label, False)
        return self.ok(label, True)

    ###########################################################################
    def ok_data_cksum(self, label, expected, data):
        """
        TAPS evaluation, hashing result of output
        """
        self.DEBUG("=> {}".format(data))
        cksum = md5(str([data]))
        if cksum != expected:
            self.DEBUG("Expected: {} != found: {}".format(expected, cksum))
            return self.ok(label, False)
        return self.ok(label, True)

    ###########################################################################
    def exit(self):
        """
        Exit appropriate for the test run, with status equal to failed tests"
        """
        self.NOTIFY("{0} failed".format(self._tap_bad))
        sys.exit(self._tap_bad)

    ###########################################################################
    def log(self, msg):
        """
        Send to tap logfile
        """
        if isinstance(msg, str):
            msg = msg.encode('utf-8')
        with open(self._tap_logfile, 'ab') as out_fd:
            out_fd.write(msg)
            out_fd.flush()

    ###########################################################################
    def log_header(self, label, filed=None):
        """
        Log, but in header formatting.
        """
        banner = "=" * 80
        banner = banner + "\n=== " + label + "\n" + banner + "\n"
        if filed:
            filed.write(banner)
        else:
            self.log(banner)

    ###########################################################################
    # pylint: disable=missing-docstring
    def inline_unit(self, *module, **kwargs):
        """
        Inline does not redirect stdout
        """
        (failed, total) = doctest.testmod(*module)
        if module:
            module = module[0]
            if isinstance(module, ModuleType):
                label = module.__name__ # pylint: disable=no-member
            else:
                label = module.__class__.__name__
        else:
            label = self.__class__.__name__
        self.ok(label + " unit tests", failed <= 0)
        if kwargs.get('exit_on_fail') and failed:
            self.NOTIFY("see test.log")
            self.exit()
        return failed, total

    ###########################################################################
    # pylint: disable=missing-docstring
    def unit(self, lib_base, modfile, exit_on_fail=False):
        failed = 1
        with open(self._tap_logfile, 'a') as out_fd:
            self.log_header(modfile + " unit test", filed=out_fd)
            sub = subprocess.Popen(['python', '-m', 'doctest',
                                    '-v', lib_base + "/" + modfile],
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
            for line in sub.stdout:
                out_fd.write(line.decode())
            for line in sub.stderr:
                out_fd.write(line.decode())
            sub.wait()
            failed = sub.returncode

        self.ok(modfile + " unit tests", failed <= 0)

        if exit_on_fail and failed:
            self.NOTIFY("see test.log")
            self.exit()

        return failed

    ###########################################################################
    # pylint: disable=missing-docstring
    def lint(self, lib_base, modfile, exit_on_fail=False):
        failed = 1
        rating = ''
        with open(self._tap_logfile, 'a') as out_fd:
            sub = subprocess.Popen(['python', '-m', 'pylint',
                                    lib_base + "/" + modfile],
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)

            self.log_header(modfile + " lint test", filed=out_fd)

            for line in sub.stdout:
                line = line.decode()
                out_fd.write(line)
                m = re.search("Your code has been rated at ([0-9.]+)/", line)
                if m:
                    rating = m.group(1)
                    if float(rating) == 10:
                        failed = 0
            for line in sub.stderr:
                out_fd.write(line.decode())
            sub.wait()

        if failed:
            with open(self._tap_logfile, 'r') as log:
                rx_ignore = re.compile(r'^(I:|==+)')
                rx_end = re.compile(r'^Report$')
                for line in log:
                    if rx_ignore.search(line):
                        continue
                    if rx_end.search(line):
                        break
                    self.NOTIFY(line.strip())

        self.ok("{} lint test ({}/10)" .format(modfile, rating), failed <= 0)

        if exit_on_fail and failed:
            self.exit()

        return failed




#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END
"""
Common repository for back-end / Reflex Engine
"""

import os
import time
import stat # chmod
import subprocess
import re
import logging
import sys
import traceback
import datetime
import dateparser
try:
    from builtins import input # pylint: disable=redefined-builtin
    get_input = input # pylint: disable=invalid-name
except: # pylint: disable=bare-except
    get_input = raw_input # pylint: disable=invalid-name, undefined-variable

import ujson
import dictlib
import rfx
import rfx.tabulate
from rfx import NotFoundError, threadlock, json4human
import rfx.client

################################################################
# helper wraps around a conf object
def getconf(cfg, *key):
    """Helper for nice error handling"""
    try:
        return _getconf(cfg, *key)
    except KeyError:
        raise ValueError("Cannot find config " + ".".join(key))

def _getconf(cfg, *key):
    """Recursive Version"""
    if len(key) == 1:
        return cfg[key[0]]
    return _getconf(cfg[key[0]], *key[1:])

def parse_dates(input_str):
    """
    Accept a string of two posix dates separated by ~,
    Return two datetime objects
    """
    if not input_str:
        return None
    split = re.split(r'\s*~\s*', input_str + "~")[0:2]
    start = dateparser.parse(split[0])
    if not start:
        raise ValueError("Invalid date: " + split[0])
    end = dateparser.parse(split[1] or 'now')
    if not end:
        raise ValueError("Invalid date: " + split[1])

    return dict(start=start.timestamp(), end=end.timestamp())

################################################################
class Engine(rfx.Base):
    """
    Reflex Engine Class - Reflex Engine
    """

    # used for speeding up repeated queries
    cache = {}
    session = None

    ###########################################################################
    def __init__(self, **kwargs):
        super(Engine, self).__init__(**kwargs)
        base = kwargs.get('base')
        if base:
            rfx.Base.__inherit__(self, base)

        # adjust requests library logging
        if self.do_DEBUG():
            logging.getLogger("requests").setLevel(logging.DEBUG)
        else:
            logging.getLogger("requests").setLevel(logging.WARNING)

        self.DEBUG("Engine.__init__(base={0})".format(base))

        if not 'REFLEX_URL' in self.cfg.keys():
            self.ABORT("Unable to find 'REFLEX_URL' in config")

        self.session = rfx.client.Session(debug=self.debug)
        rfx.Base.__inherit__(self.session, base)

    ###########################################################################
    # pylint: disable=unused-argument
    def list_objects(self, obj_type, apikey=None):
        """Uses build-in environmental connection information"""
        return self.session.list(obj_type)

    ###########################################################################
    # pylint: disable=unused-argument,too-many-arguments
    def create_object(self, obj_type, payload, reason=None, apikey=None, notify=True):
        """Uses build-in environmental connection information"""
        return self.session.create(obj_type, payload)

    ###########################################################################
    def delete_object(self, obj_type, obj_target, apikey=None, notify=True):
        """Uses build-in environmental connection information"""
        return self.session.delete(obj_type, obj_target)

    ###########################################################################
    # pylint: disable=too-many-arguments
    def delta_update_object(self, obj_type, obj_target, payload, **kwargs):
        """Wraps update_object and makes a dict_union change"""
        return self.session.patch(obj_type, obj_target, payload)

    ###########################################################################
    # pylint: disable=too-many-arguments
    def update_object(self, obj_type, obj_target, payload, reason=None, apikey=None, notify=True):
        """Uses build-in environmental connection information"""
        return self.session.update(obj_type, obj_target, payload)

    ###########################################################################
    def get_object(self, obj_type, obj_target, **kwargs): #apikey=None, notify=True):
        """Uses build-in environmental connection information"""
        if kwargs.get('raise_error') is True:
            return self.session.get(obj_type, obj_target)
        else:
            try:
                return self.session.get(obj_type, obj_target)
            except: # pylint: disable=bare-except
                return None

    ###########################################################################
    @threadlock
    def cache_get_object(self, obj_type, obj_target, **kwargs):
        """
        Cache wrapper around Engine.get_object, return None on failed lookup.
        Optionally re-enable error throwing
        """
        raise_error = kwargs.get('raise_error', True)
        if 'raise_error' in kwargs:
            del kwargs['raise_error']
        if obj_type in self.cache:
            if obj_target in self.cache[obj_type]:
                return self.cache[obj_type][obj_target]
        else:
            self.cache[obj_type] = {}
        try:
            obj = self.get_object(obj_type, obj_target, **kwargs)
        except NotFoundError:
            self.cache[obj_type][obj_target] = None
            if raise_error:
                raise
            return None
        self.cache[obj_type][obj_target] = obj
        return obj

    ###########################################################################
    @threadlock
    def cache_update_object(self, obj_type, obj_target, payload, **kwargs):
        """
        Cache wrapper around Engine.update_object, invalidates cache for that object
        """
        value = self.update_object(obj_type, obj_target, payload, **kwargs)
        if value:
            if obj_type in self.cache:
                if obj_target in self.cache[obj_type]:
                    del self.cache[obj_type][obj_target]
        return value

    ###########################################################################
    @threadlock
    def cache_list_objects(self, obj_type, **kwargs):
        """
        Cache wrapper around Engine.list_object.
        """
        if not obj_type in self.cache:
            self.cache[obj_type] = {}

        objs = self.list_objects(obj_type, **kwargs)
        for obj in objs:
            oname = obj.get('name')
            if oname:
                self.cache[obj_type][oname] = obj

        return objs

    ###########################################################################
    @threadlock
    def cache_reset(self):
        """Clear the cache"""
        self.cache = {}

###############################################################################
def _valid_name(obj_name):
    if not obj_name:
        obj_name = ''
    if re.search(r'[^a-z0-9_.-]', obj_name, flags=re.IGNORECASE):
        return "Invalid name ({}), May only contain a-z0-9_-" \
               .format(obj_name)
    return None

###############################################################################
class EngineCli(rfx.Base):
    """Basic CLI for Engine Class"""

    ###########################################################################
    # pylint: disable=super-init-not-called
    def __init__(self, base=None):
        if base:
            rfx.Base.__inherit__(self, base)
        self.rcs = rfx.client.Session(debug=self.debug, base=self)

    ###########################################################################
    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    def list_cli(self, obj_type, parsed, argv):
        """list an object.  see --help"""
        try:
            archive = parse_dates(parsed.get('--archive'))
        except ValueError as err:
            self.ABORT(str(err))

        # if it is a real regex, we must reduce scope on our side
        filter_re = None
        name = None
        if argv and argv[0]:
            if re.search(r'[^a-z0-9.-]', argv[0]):
                try:
                    filter_re = re.compile(argv[0])
                except Exception as err: # pylint: disable=broad-except
                    self.ABORT("Invalid regular expression: {}\n\t{}".format(argv[0], str(err)))
            else:
                name = argv[0]

        limit_expr = None
        if parsed.get('--expression'):
            limit_expr = parsed['--expression']

        stdout = list()
        stderr = list()
        show = list() # uppercase header
        if parsed.get('--stderr'):
            stderr = [x.lower() for x in re.split(r"\s*,\s*", parsed['--stderr'])]
        if parsed.get('--stdout'):
            stdout = [x.lower() for x in re.split(r"\s*,\s*", parsed['--stdout'])]

        if parsed.get('--show'):
            show = [x.lower() for x in re.split(r"\s*,\s*", parsed['--show'])]
            if not stderr and not stdout:
                stderr = [x for x in range(0, len(show)) if show[x] != "name"]
        elif stderr or stdout:
            show = stdout + stderr

        if not show:
            show = ['name', 'id']
            if archive:
                show += ['updated_at']

        cols = show.copy()
        ncols = len(cols)

        # find out any sub.key references -- we can only request the top level
        subs = dict()
        for pos in range(len(cols)): # pylint: disable=consider-using-enumerate
            if '.' in cols[pos]:
                elem = cols[pos]
                top = elem[0:elem.index('.')]
                sub = elem[elem.index('.')+1:]
                subs[pos] = sub
                cols[pos] = top

        cols = [x.lower() for x in cols]
        results = [[x.upper() for x in show]]

        try:
            objs = self.rcs.list(obj_type, cols=cols, match=name, archive=archive)
        except rfx.client.ClientError as err:
            self.ABORT(str(err))

        for obj in objs:
            if filter_re and not filter_re.search(obj['name']):
                continue
            if limit_expr:
                try:
                    context = {'obj': dictlib.Obj(**obj), 'True': True, 'False': False}
                    # pylint: disable=eval-used
                    if not eval(limit_expr, {'__builtins__':{}, 'rx': re}, context):
                        continue
                except KeyError:
                    continue
                except SyntaxError:
                    self.ABORT(traceback.format_exc(0))
                except: # pylint: disable=bare-except
                    self.ABORT(traceback.format_exc())

            # make it insensitive
            obj_lower = dict()
            for key, value in obj.items():
                obj_lower[key.lower()] = value

            row = list()
            for key_x in range(0, ncols):
                if cols[key_x] == 'updated_at':
                    posix_time = int(obj_lower[cols[key_x]])
                    val = "{} ({})".format(posix_time,
                                           datetime.datetime.fromtimestamp(posix_time))
                    row.append(val)
                else:
                    try:
                        if subs and subs.get(key_x):
                            row.append(dictlib.dig(obj_lower[cols[key_x]], subs[key_x]))
                        else:
                            row.append(obj_lower[cols[key_x]])
                    except (KeyError, TypeError):
                        row.append('')
            results.append(row)

        rfx.tabulate.cols(results, stderr=stderr, header=True, fmt=parsed.get('--format', 'txt'))

    ###########################################################################
    def copy_cli(self, obj_type, obj_source, obj_dest):
        """copy an object.  see --help"""
        try:
            self.rcs.get(obj_type, obj_dest)
            self.ABORT("Destination object '{0}' already exists!".format(obj_dest))
        except rfx.client.ClientError:
            pass

        try:
            obj = self.rcs.get(obj_type, obj_source)
        except rfx.client.ClientError as err:
            self.ABORT(str(err))

        if 'id' in obj:
            del obj['id']

        self.OUTPUT("Copy {0} {1} to {2}".format(obj_type, obj_source, obj_dest))
        obj['name'] = obj_dest
        parsed = {'--content': json4human(obj), 'name': obj_dest}
        return self.edit_cli(obj_type, parsed, [], pause=False)

    ###########################################################################
    def get_cli(self, obj_type, parsed, argv):
        """get an object.  see --help"""
        try:
            archive = parse_dates(parsed.get('--archive'))
        except ValueError as err:
            self.ABORT(str(err))
        obj_name = parsed['name']
        try:
            obj = self.rcs.get(obj_type, obj_name, archive=archive)
            if argv and argv[0]:
                key = argv[0]
                if key[:4] == 'obj.':
                    key = key[4:]
                value = dictlib.dig_get(obj, key, '')
                if parsed.get('--format') == 'txt':
                    if isinstance(value, str):
                        self.OUTPUT(value)
                    elif isinstance(value, list):
                        self.OUTPUT(", ".join(value))
                    else:
                        self.OUTPUT(json4human(value))
                else:
                    self.OUTPUT(json4human(value))
            else:
                self.OUTPUT(json4human(obj))
        except rfx.client.ClientError as err:
            #self.ABORT("Exception: " + traceback.format_exc())
            self.ABORT(str(err))
        except NotFoundError:
            self.ABORT("Cannot find object '" + obj_name + "'")
        except Exception as err: # pylint: disable=broad-except
            self.DEBUG("Exception: " + str(err))
            self.ABORT("Exception: " + traceback.format_exc())

    ###########################################################################
    # pylint: disable=unused-argument
    def create_cli(self, obj_type, parsed, argv):
        """create an object.  see --help"""
        obj_name = parsed['name']
        content = self._get_default(obj_type, obj_name, parsed.get('--content'))
        self.rcs.update(obj_type, obj_name, content)
        self.get_cli(obj_type, parsed, [])

    ###########################################################################
    # pylint: disable=unused-argument
    def delete_cli(self, obj_type, parsed, argv):
        """delete an object.  see --help"""
        obj_name = parsed['name']
        try:
            result = self.rcs.delete(obj_type, obj_name)
            if result['status'] == 'deleted':
                self.NOTIFY("Deleted " + obj_type.capitalize() + " '" + obj_name + "'")
            else:
                self.NOTIFY("Unable to delete: {}".format(result))
        except Exception as err: # pylint: disable=broad-except
            self.ABORT("Cannot delete object '{0}': {1}".format(obj_name, err))

    ###########################################################################
    # pylint: disable=unused-argument
    def edit_cli(self, obj_type, parsed, argv, pause=True):
        """edit a commplete object.  see --help"""
        obj_name = parsed['name']
        msg = _valid_name(obj_name)
        if msg:
            self.ABORT(msg)

        #if re.search("[^a-z0-9_-]", obj_name, flags=re.IGNORECASE):
        #    self.ABORT("Invalid {0} name ({1}), May only contain a-z0-9_-"
        #               .format(obj_type, obj_name))
        scratchdir = os.path.expanduser("~") + "/.rfx.scratch"
        if not os.path.exists(scratchdir):
            os.mkdir(scratchdir)
        os.chmod(scratchdir, stat.S_IRWXU)
        localfile = scratchdir + "/" + obj_type + "." + obj_name + ".json"

        try:
            content = self.rcs.get(obj_type, obj_name)
            for key in ('updated_at', 'updated_by', 'created_at'):
                if key in content:
                    del content[key]
            with open(localfile, 'wt') as write_fd:
                write_fd.write(ujson.dumps(content,
                                           indent=2,
                                           sort_keys=True,
                                           escape_forward_slashes=False))

        except Exception as err: # pylint: disable=broad-except
            self.DEBUG("Exception: " + str(err))
            if str(err) == "Forbidden":
                self.ABORT(str(err))
            self.NOTIFY("Object '{0}' does not exist, creating new...".format(obj_name))
            if pause:
                time.sleep(1)
            with open(localfile, 'w') as write_fd:
                data = self._get_default(obj_type,
                                         obj_name,
                                         parsed.get('--content'),
                                         read=False)
                write_fd.write(ujson.dumps(data,
                                           indent=2,
                                           sort_keys=True,
                                           escape_forward_slashes=False))

        data = None
        while data is None:
            try:
                editor = os.environ['EDITOR']
            except: # pylint: disable=bare-except
                editor = 'vim'

            try:
                subprocess.call([editor, localfile])
            except OSError as err:
                self.ABORT("Unable to run editor '{}':\n\t{}"
                           .format(editor, err))

            with open(localfile, 'r') as read_fd:
                data = self._json_load_handle_error(read_fd.read(),
                                                    ask2continue=True,
                                                    cleanfile=localfile)
        os.unlink(localfile)

        answer = get_input("Commit changes? [y] ")

        # incase of a rename
        data_name = data.get('name', None)
        if data_name is None:
            data['name'] = obj_name
        else:
            obj_name = data_name

        if not answer or re.match("^(yes|y)$", answer, flags=re.IGNORECASE):
            try:
                self.rcs.update(obj_type, obj_name, data)
            except rfx.client.ClientError as err:
                self.ABORT(str(err))

    ###########################################################################
    # pylint: disable=unused-argument
    def update_cli(self, obj_type, parsed, argv):
        """update a commplete object.  see --help"""
        obj_name = parsed['name']
        data = self._load_content(obj_name, parsed.get('--content'))
        if data:
            self.rcs.update(obj_type, obj_name, data)
        else:
            self.ABORT("No content to update?")

    ###########################################################################
    # pylint: disable=unused-argument
    def slice_cli(self, obj_type, parsed, argv):
        """slice limit-expression extract-keys (comma delimited). see --help"""

        name_rx = re.compile('^' + parsed['name-filter'])
        limit = parsed['limit-expression']
        extract = parsed['key']
        if extract[:4] == "obj.":
            extract = extract[4:]

        extracted = set()
        extracted_objs = dict()

        for obj in Engine(base=self).session.list(obj_type, cols='*'):
            if not name_rx.search(obj['name']):
                continue
            context = {'obj': dictlib.Obj(**obj), 'True': True, 'False': False}
            try:
                # pylint: disable=eval-used
                if not eval(limit, {'__builtins__':{}, 'rx': re}, context):
                    continue
            except KeyError:
                print("key error")
                continue
            except: # pylint: disable=bare-except
                traceback.print_exc()

            value = dictlib.dig_get(obj, extract)
            if not value:
                continue

            if isinstance(value, list):
                extracted = extracted.union(value)
            else:
                extracted.add(value)

            extracted_objs[obj['name']] = value

        if parsed.get('--format', 'txt') == 'txt':
            self.OUTPUT(" ".join(extracted) + "\n")
        if parsed.get('--format', 'txt') == 'list':
            # pylint: disable=consider-iterating-dictionary
            for obj in sorted(extracted_objs.keys()):
                self.OUTPUT("{}: {}".format(obj, extracted_objs[obj]))
        else:
            self.OUTPUT(ujson.dumps(extracted) + "\n")

    ###########################################################################
    # pylint: disable=unused-argument
    def merge_cli(self, obj_type, parsed, argv):
        """merge data into a current object.  see --help"""
        obj_name = parsed['name']
        data = self._load_content(obj_name, parsed.get('--content'))
        try:
            data = self.union_dict(self.rcs.get(obj_type, obj_name), data)
        except Exception as err: # pylint: disable=broad-except
            self.DEBUG("Excpeption: " + str(err))
            self.ABORT("Cannot find object '" + obj_name + "'")

        self.rcs.update(obj_type, obj_name, data)

    ###########################################################################
    def _json_load_handle_error(self, content, ask2continue=False, cleanfile=None):
        try:
            data = ujson.loads(content)
            msg = _valid_name(data.get('name'))
            if msg:
                raise ValueError(msg)
            return data
        except ValueError as err:
            if content:
                linenbr = 1
                for line in content.split("\n"):
                    self.NOTIFY("{0:3}: {1}".format(linenbr, line))
                    linenbr += 1
            self.NOTIFY("Unable to parse: " + str(err))
            if ask2continue:
                answer = get_input("Return to editor? [y] ")
                if not re.match("^(no|n)$", answer, flags=re.IGNORECASE):
                    return None
            if cleanfile:
                os.unlink(cleanfile)
            self.ABORT("Cannot continue.")

    ###########################################################################
    # pylint: disable=unused-argument
    def _load_content(self, obj_name, obj_content, read=True):
        if not obj_content and read:
            self.NOTIFY("Reading JSON data from stdin...")
            obj_content = ''
            for line in sys.stdin.readlines():
                obj_content += line
        if obj_content:
            return self._json_load_handle_error(obj_content)
        return {}

    ###########################################################################
    # NOTE: merge w/newapp / populate code
    def _get_default(self, obj_type, obj_name, obj_content, read=True):
        try:
            data = getattr(self, "_get_default_" + obj_type)(obj_name)
        except: # pylint: disable=bare-except
            data = {}

        content = self._load_content(obj_name, obj_content, read=read)
        if content:
            data = self.union_dict(data, content)
        return data

################################################################################
class Release(rfx.Base):
    """Records of the software build/release/package process"""

    obj = None
    name = ''
    engine = None
    statuses = {
        "prep": "skipped",
        "compile": "skipped",
        "test-unit": "incomplete",
        "assemble": "incomplete",
        "test-integration": "incomplete",
        "test-run": "skipped",
        "test-functional": "skipped",
        "finish": "skipped"
    }
    status_vals = {
        "incomplete": 1,
        "started": 1,
        "success": 1,
        "failure": 1,
        "skipped": 1
    }

    ############################################################################
    def __init__(self, *args, **kwargs):
        super(Release, self).__init__(*args, **kwargs)
        if 'base' in kwargs:
            rfx.Base.__inherit__(self, kwargs['base'])
        self.engine = Engine(base=self)

    ############################################################################
    def can_edit(self, app, version, args):
        """check if the object is changeable or not, ABORT if not"""

        self.load(app, version)

        if not self.obj:
            self.obj = {
                'name': self.name,
                'application': app,
                'version': version,
                'type': args.type or 'deploy-pkg',
                'status': {},
                'state': 'pending'
            }

        if self.obj.get('state', 'pending') in ('failed', 'ok'):
            self.ABORT("Release is complete, cannot be changed")

    ############################################################################
    def load(self, app, version, abort=False):
        """Pull the described app/version onto the object"""
        if not self.name:
            self.name = (app + '-' + version).replace("_", "-").replace('.', '-')

        if not self.obj:
            try:
                self.obj = self.engine.get_object("release", self.name, notify=False)
            except rfx.NotFoundError as err:
                if abort:
                    self.ABORT(str(err[0])) # pylint: disable=unsubscriptable-object

    ############################################################################
    def update(self):
        """push an update of the object"""
        if self.obj:
            self.engine.update_object('release', self.name, self.obj, notify=False)
        else:
            self.ABORT("No object to update")

    ############################################################################
    def status_cli(self, app, version, args):
        """report or change the status of a release"""
        if not version:
            if args.setkey:
                self.ABORT("Cannot set a status without a version")

            # NOTE: not efficient, fix Reflex Engine itself and come fix this
            re_rel = re.compile("^" + app)
            count = 0
            for obj in self.engine.list_objects('release'):
                if not re_rel.search(obj['name']):
                    continue
                count += 1
                app = obj.get('application')
                version = obj.get('version')
                state = obj.get('state')
                self.NOTIFY("{:30} {:9} {}".format(app, version, state))
            self.NOTIFY("Total: {}".format(count))
            return

        if args.setkey:
            self.can_edit(app, version, args)
            # parse setkeys, update
            setkeys = []
            errors = 0
            for setkey in args.setkey:
                key, val = (setkey.lower() + "=").split("=")[:2]
                if not key in self.statuses:
                    self.NOTIFY("Invalid status '" + key + "', not one of:\n\t" +
                                ", ".join(self.statuses.keys()))
                    errors += 1
                elif not val in self.status_vals:
                    self.NOTIFY("Invalid status value '{}' for '{}', not one of:\n\t{}"
                                .format(val, key, ", ".join(self.status_vals.keys())))
                    errors += 1
                else:
                    setkeys.append([key, val])
            if errors:
                self.ABORT("Cannot continue")
            if setkeys:
                for key, val in setkeys:
                    self.obj['status'][key] = val
                    self.NOTIFY("set {}={}".format(key, val))
                self.update()
                return

        self.load(app, version, abort=True)
        self.OUTPUT(ujson.dumps(self.obj, indent=2, escape_forward_slashes=False))

    ############################################################################
    def link_cli(self, app, version, args):
        """link to a release"""
        self.can_edit(app, version, args)

        # could parseurl if we wanted...
        self.obj['link'] = args.link
        self.update()
        self.NOTIFY("link=" + args.link)

    ############################################################################
    def data_cli(self, app, version, args):
        """get or specify metadata for a release"""

        data = args.data
        if data == '-':
            self.NOTIFY("Reading JSON data from stdin...")
            data = sys.stdin.read()
        try:
            data = ujson.loads(data)
        except ValueError as err:
            self.ABORT("Unable to parse: " + str(err))

        self.can_edit(app, version, args)

        # could parseurl if we wanted...
        self.obj['data'] = data
        self.update()
        self.NOTIFY("data=" + ujson.dumps(data, indent=2, escape_forward_slashes=False))

    ############################################################################
    def finish_cli(self, app, version, args):
        """finish a release"""

        self.can_edit(app, version, args)
        status = self.obj.get('status', {})
        state = 'ok'
        for status_name in status:
            default = self.statuses.get(status_name)
            if status.get(status_name, default) not in ('ok', 'skipped'):
                state = 'failed'
                break

        self.obj['state'] = state
        self.update()

    ############################################################################
    # pylint: disable=unused-argument
    def delete_cli(self, app, version, args):
        """delete a release"""

        self.load(app, version)
        if self.obj:
            self.engine.delete_object('release', self.name)




#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END

"""
Launch Control
"""

import os
import traceback
import socket
import dictlib
import rfx
from rfx.config import ConfigProcessor
from rfx import json4human, json4store
from rfx.backend import Engine
from rfx.action import Action

################################################################################
# pylint: disable=too-many-instance-attributes
class App(rfx.Base):
    """Manage configuration files for containers"""

    launch_service = None
    launch_pipeline = None
    launch_config = None
    launch_action = None
    launch_target = ''
    launch_env = None
    launch_exec = None
    launch_rundir = ''
    launch_cfgdir = ''
    launch_type = 'exec'
    launch_peers = None

    ############################################################################
    # pylint: disable=super-init-not-called
    def __init__(self, base=None):
        rfx.Base.__inherit__(self, base)
        self.rcs = rfx.client.Session(base=base)
        self.get_my_nameip()
        self.launch_peers = dict()
        self.launch_exec = list()
        self.launch_env = dict()
        self.launch_pipeline = dict()
        self.launch_service = dict()

    ############################################################################
    def __repr__(self):
        return json4store(self.export()) #, indent=2)

    def export(self):
        """Return a complete object"""
        obj = {}

        if self.launch_pipeline:
            obj = dictlib.union(obj, self.launch_pipeline)
            obj['pipeline'] = obj['name']

        if self.launch_service:
            obj = dictlib.union(obj, self.launch_service)

        if self.launch_config:
            obj = dictlib.union(obj, self.launch_config) #.export())

        obj.pop('id', '')
        obj.pop('name', '')
        obj.pop('createdAt', '')
        obj.pop('updatedAt', '')

        return obj

    ############################################################################
    def launch_service_prep(self, name, commit=True):
        """Prepare a service for launching"""
        name_parts = name.split(":")
        action = None
        service = name_parts[0]
        if len(name_parts) > 1:
            action = name_parts[1]
        try:
            self.launch_service = self.rcs.get('service', service)
        except rfx.client.ClientError as err:
            if "Endpoint or object not found" in str(err):
                self.ABORT("Cannot find service: " + service)
            self.ABORT("Cannot continue: " + str(err))

        # xODO: Add trap and error handling if pipeline privs fail
        self.launch_pipeline = self.rcs.get('pipeline',
                                            self.launch_service['pipeline'])
        os.environ["APP_PIPELINE"] = self.launch_pipeline['name']
        os.environ["APP_SERVICE"] = self.launch_service['name']

        # pull cfg from service, before doing an action.. which may override it
        def setcfg(envkey, cfgkey):
            """shortcut"""
            launch = self.launch_pipeline.get('launch', {})
            if not envkey in os.environ and cfgkey in launch:
                value = self.sed_env(launch[cfgkey], {}, '')
                setattr(self, cfgkey, value)

        setcfg('APP_CFG_BASE', 'cfgdir')
        setcfg('APP_RUN_BASE', 'rundir')

        self.launch_type = self.launch_pipeline.get('launch', {}).get('type', 'exec')
        if action == 'none': # special case
            self._launch_prep_action(action, commit=commit)
            action = None

        elif action or self.launch_type == 'action':
            self.launch_type = 'action'
            self._launch_prep_action(action, commit=commit)

        else:
            self._launch_prep_exec(service, commit=commit)

        # for security, do not expose this to the running app
        for key in ['REFLEX_APIKEY']:
            if key in os.environ:
                del os.environ[key]
            if action and key in self.launch_action.env:
                del self.launch_action.env[key]

        return self

    ############################################################################
    # pylint: disable=unused-argument
    def _load_reflex_engine_config(self, target, commit=False):
        """get a from Reflex Engine"""
        cproc = ConfigProcessor(base=self, rcs=self.rcs, peers=self.launch_peers)
        conf = cproc.flatten(self.launch_service['config'])
        if commit:
            cproc.commit(conf, dest=self.launch_cfgdir)
        else: # normally called by commit
            cproc.prep_setenv(conf, dest=self.launch_cfgdir)
        conf = cproc.prune(conf)
        self.launch_config = conf
        return conf

    ############################################################################
    def _launch_prep_action(self, action_target, commit=True):
        """
        With launch control, local immutable package env configs take precendence
        over remotely defined (Reflex Engine) configs.
        Important: do not set secrets or environmentally unique information in
        local environment--that is only for things relevant to the local immutable
        package, such as the version of the stack required (java/node/etc).
        """

        action = Action(base=self, extcfg=self, colorize=False)

        # set APP_CFG_BASE and APP_RUN_DIR first -- these are not in
        # setenv because they must exist prior to setenv (which can have macro subs)
        local_config = action.config.get('config', {})

        # pylint: disable=unused-argument
        def get_conf(name):
            """Pull some configs based on local first, then service"""
            if name in local_config:
                return local_config[name]
            elif name in self.launch_pipeline.get('launch', {}):
                return self.launch_pipeline['launch'][name]
            raise ValueError("Unable to find '" + name + "' in local or pipeline definition")

        os.environ['APP_CFG_BASE'] = self.sed_env(get_conf('cfgdir'), {}, '')
        self.launch_cfgdir = os.environ['APP_CFG_BASE']
        os.environ['APP_RUN_BASE'] = self.sed_env(get_conf('rundir'), {}, '')
        self.launch_rundir = os.environ['APP_RUN_BASE']

        # Load Reflex Engine config, after ^^ environ changes
        self._load_reflex_engine_config(self.launch_service['config'], commit=commit)

        # load action configs.  pull Reflex Engine config expanded, merge with action config
        # and redo macro expansion
        conf = self.launch_config
        conf.setenv = dictlib.union(conf.setenv,
                                    action.config.get('setenv', {}))
        cproc = ConfigProcessor(base=self, rcs=self.rcs)
        conf.setenv = cproc.macro_expand_dict(conf.setenv)
        for key, value in conf.setenv.items():
            value = self.sed_env(str(value), local_config, '')
            conf.setenv[key] = value
            os.environ[key] = value
            action.env[key] = value

        if action_target == 'none': # special case
            return self
        elif action_target:
            self.launch_target = action_target
        elif self.launch_pipeline.get('launch', {}).get('target', None):
            self.launch_target = self.launch_pipeline['launch']['target']
        else:
            raise ValueError("No launch service action target (svc:target)")

        action.verify(self.launch_target)
        self.launch_action = action

        return self

    ############################################################################
    def _launch_prep_exec(self, name, commit=True):
        launch = self.launch_pipeline.get('launch', {})
        self.launch_rundir = self.sed_env(launch.get('rundir', '.'), {}, '')
        self.launch_cfgdir = self.sed_env(launch.get('cfgdir', '.'), {}, '')
        self.launch_exec = [self.sed_env(elem, {}, '') for elem in launch.get('exec', [])]

        os.environ["APP_RUN_BASE"] = self.launch_rundir
        os.environ["APP_CFG_BASE"] = self.launch_cfgdir
        try:
            self._load_peers()
        except rfx.client.ClientError as err:
            self.NOTIFY("Unable to load peers: " + str(err))

        self._load_reflex_engine_config(self.launch_service['config'], commit=commit)

        return self

    ############################################################################
    def _launch_action(self, name):
        """
        Launch a service using the 'action' means, where the launch
        data is stored within the immutable package as an action.
        """

        self._launch_update_instance()
        self.launch_action.do(self.launch_target)

    ############################################################################
    def _launch_exec(self, name):
        """
        Launch a service using the 'exec' means, where the launch
        data is stored on the pipeline object.
        """
        if not os.path.isdir(self.launch_rundir):
            self.ABORT("Unable to find launch rundir (" +
                       self.launch_rundir + "), cannot launch!")
        os.chdir(self.launch_rundir)
        if not os.path.isfile(self.launch_exec[0]):
            self.NOTIFY("Unable to find launch program: " +
                        self.launch_exec[0])
        if self.do_DEBUG():
            msg = "Launch Env:\n"
            for key in self.launch_config.setenv.keys():
                msg += "  {}={}".format(key, self.launch_config.setenv[key])
            self.NOTIFY(msg) # keep it as one 'message' for log handlers
        self.NOTIFY("Launch working directory=" + self.launch_rundir)
        self.NOTIFY("Launch exec=['" + "','".join(self.launch_exec) + "']")
        self._launch_update_instance()
        os.execv(self.launch_exec[0], self.launch_exec)

    ############################################################################
    def _load_peers(self):
        """Query Reflex engine for peers to self"""

        self.launch_peers = dict(ip0=dict(), ip1=dict())
        for inst in self.rcs.list('instance',
                                  match=self.launch_service['name'],
                                  cols=['address', 'status', 'name'],
                                  raise_error=False):
            if inst['name'] == self.my_host:
                continue
            for iplabel in self.launch_peers:
                if inst.get('status', 'failed') != 'ok':
                    continue
                ipnbr = inst['address'].get(iplabel, '')
                if ipnbr:
                    self.launch_peers[iplabel][inst['name']] = ipnbr

    ############################################################################
    def _launch_update_instance(self):
        try:
            myname = socket.gethostname()
            myips = socket.gethostbyname_ex(myname)[2]
            addrs = {}
            for idx, ipnbr in enumerate(myips):
                addrs["ip" + str(idx)] = ipnbr
            self.NOTIFY("Instance Ping " + myname)
            self.rcs.instance_ping(myname, {
                "status": "ok",
                "service": self.launch_service['name'],
                "name": myname,
                "address": addrs
            })
        except Exception as err: # pylint: disable=broad-except
            if self.do_DEBUG():
                self.NOTIFY("Unable to update instance: " + str(err))
                self.DEBUG(traceback.format_exc())
            else:
                self.NOTIFY("Unable to update instance: " + str(err))

################################################################################
class LaunchCli(App):
    """CLI interface to Launch"""

    ############################################################################
    # pylint: disable=inconsistent-return-statements
    def get_target(self, *argv):
        """
        Pull the target from either the first arg, environment, or abort.
        """
        env_service = os.environ.get("REFLEX_SERVICE")
        if argv and argv[0]:
            return argv[0]
        elif not env_service:
            self.ABORT("No target defined (either as argument or set as REFLEX_SERVICE)")
        else:
            return env_service

    ############################################################################
    # pylint: disable=unused-argument
    def env_cli(self, argv, args, cli):
        """
        Print bash friendly environment declarations.
        Intended to be used in a manner like:
        ```eval $(launch env name)```
        """

        name = self.get_target(*argv)
        export = "export "
        if args.get('--noexport', False):
            export = ""

        try:
            self.launch_service_prep(name, commit=False)
            for key in sorted(self.launch_config.setenv): #_expanded:
                value = self.launch_config.setenv[key]
                if export:
                    value = json4store(value)
                self.OUTPUT('{}{}={}'.format(export, key, value))
        except rfx.NotFoundError:
            self.ABORT(traceback.format_exc(0))

    ############################################################################
    def service_cli(self, argv, args, cli):
        """
        Launch a service, after preparing the config, replacing current
        process with new process (exec).
        Supports two types: exec and action
        """
        service = self.get_target(*argv)
        debug = self.do_DEBUG()
        rfx.unbuffer(self.notifyfd)
        rfx.unbuffer(self.outputfd)
        try:
            self.launch_service_prep(service, commit=True)
            for key in sorted(self.launch_config.setenv): # .items():
                value = self.launch_config.setenv[key]
                os.environ[key] = value
                if debug:
                    self.NOTIFY("export {}={}".format(key, value))
            getattr(self, "_launch_" + self.launch_type)(service)

        except Exception as err: # pylint: disable=broad-except
            self.DEBUG(traceback.format_exc())
            self.ABORT("Unable to launch " + service + ": " + str(err))

    ############################################################################
    def config_cli(self, argv, args, cli):
        """Show just our config"""
        target = self.get_target(*argv)
        try:
            self.launch_service_prep(target, commit=args.get('--commit', False))
            self.OUTPUT(json4human(self.export()))
        except (ValueError, Exception): # pylint: disable=broad-except
            self.DEBUG(traceback.format_exc())
            self.ABORT(traceback.format_exc(0))



#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END

"""Encryption Wrapper"""

import base64
import nacl.secret
import nacl.utils

################################################################################
# pylint: disable=too-few-public-methods
class Key(object):
    """
    Key wrapper

    >>> Key("MmRnFLmmF4iqXxKAhjkrkC/+pdABVQipeKSv2EZAqOY=").encode()
    'MmRnFLmmF4iqXxKAhjkrkC/+pdABVQipeKSv2EZAqOY='
    """

    value = b''

    def __init__(self, *existing):
        if existing:
            self.value = base64.b64decode(existing[0])
        else:
            self.value = nacl.utils.random(nacl.secret.SecretBox.KEY_SIZE)

    def encode(self):
        """Return key encoded"""
        return base64.b64encode(self.value).decode()

################################################################################
class Cipher(object):
    """
    Generic Cipher Wrapper

    >>> key = Key('MmRnFLmmF4iqXxKAhjkrkC/+pdABVQipeKSv2EZAqOY=')
    >>> result = Cipher(key).key_encrypt("test", raw=True)
    >>> Cipher(key).key_decrypt(result, raw=True)
    'test'
    """

    key = None

    ############################################################################
    def __init__(self, key):
        """Create a secret key"""
        if isinstance(key, Key):
            self.key = key
        else:
            raise ValueError("argument is not a Key object")

    ############################################################################
    def key_encrypt(self, data, raw=False):
        """
        Encrypt data.
        Note: insert 'struct' to pull the first byte of data out as a "version"
        """
        nonce = nacl.utils.random(nacl.secret.SecretBox.NONCE_SIZE)
        encrypted = nacl.secret.SecretBox(self.key.value).encrypt(data.encode(), nonce)
        if raw:
            return encrypted
        return base64.b64encode(encrypted).decode()

    ############################################################################
    def key_decrypt(self, data, raw=False):
        """Decrypt data"""
        if not raw:
            data = base64.b64decode(data)
        return nacl.secret.SecretBox(self.key.value).decrypt(data).decode()

    # future: add pki_encrypt/decrypt



#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END

"""
Deploy Control
"""

import os
#import subprocess
#import traceback
#import sys
import socket # gethostbyname
#import ujson as json

import rfx
from rfx.backend import Engine

############################################################################
# heuristically determint our node name
def my_instance_name():
    """docstr"""
    # os environ trumps anything else; but this is not reflected by
    # how launch control handles it...
    port = os.environ.get('PORT')
    #if not port:
        # lookup on config
        # is cfg->sensitive->parameters->PORT or cfg->setenv->PORT defined?

    name = socket.gethostname().split(".")[0]
    if port:
        name += "." + str(port)

    return name

################################################################################
class Deploy(rfx.Base):
    """
    Manage Deployment
    """
    engine = None

    ############################################################################
    def __init__(self, *args, **kwargs):
        super(Deploy, self).__init__(*args, **kwargs)
        if 'base' in kwargs: # sorry, we need the base info
            rfx.Base.__inherit__(self, kwargs['base'])
        self.engine = Engine(base=self)

    ############################################################################
    def engine_get(self, otype, oname):
        """helper function"""
        return self.engine.cache_get_object(otype, oname, notify=False)

    def engine_put(self, otype, oname, obj):
        """helper function"""
        return self.engine.cache_update_object(otype, oname, obj, notify=False)


    ############################################################################
    def target_cli(self, service, args):
        """
        if version specified, set it as the target
        otherwise print existing target
        """
        obj = self.engine_get("service", service)
        if args.version:
            # FUTURE: Check for valid Build object first
            ver = args.version.strip()

            obj['target'] = ver
            self.engine_put("service", service, obj)

            # verify and bypass cache
            obj = self.engine_get("service", service) # bypass cache

            if obj['target'] != ver:
                self.ABORT("Unable to set target version?  Bad error.")

            self.NOTIFY("set target {} = {}".format(service, args.version))
        else:
            self.OUTPUT("target {} = {}".format(service, obj.get('target')))

    ############################################################################
    # pylint: disable=unused-argument
    def status_cli(self, service, args):
        """
        report instance states
        if --verify specified, deeply query each instance to verify
        """
        obj = self.engine_get("service", service)
        target = obj.get('target', '')
        nodes = set(obj.get('instances', [])).union(obj.get('active-instances'))
        harmony = True
        self.OUTPUT("target {} = {}".format(service, target))
        for node_name in nodes:
            node = self.engine_get("instance", node_name)
            vers = node.get('version', None)
            if vers != target:
                harmony = False
            self.OUTPUT("version {} = {}".format(node.get('name'), vers))

        if not harmony:
            self.ABORT("Service is not in harmony")

    ############################################################################
    # pylint: disable=unused-argument
    def update_cli(self, service, args):
        """
        Make local version match target version and restart (if necessary)
        """
        svc = self.engine_get("service", service)
        pipe = self.engine_get("pipeline", svc.get("pipeline"))
        region = pipe.get("region", None)
        if not region:
            self.ABORT("service {} -> pipeline {} has no region defined!"
                       .format(svc, pipe))
        cfg = self.engine_get("config", "ops-deploy").get("config", {}).get(region, None)
        if not cfg:
            self.ABORT("region {} for service {} is not defined on ops-deploy!"
                       .format(region, svc))

#        node_name = svc.get('name') + '.' + hostname(base) + port
        #node = self.engine_get("instance",

        # what type?  Should only be run on server based objects
        # cfg-build should use Build object... perhaps cfg object for now
        # get path
        # boto pull object
        # down service in Engine
        # -- remove nginx
        # -- verify no traffic
        # pre-activate
        # unroll
        # post-activate
        # up service in Engine
        # -- add nginx

#        - query Engine to identify target local service should be at
#        -- if it is not matching, then make it match and update instance status
#        -- instance may be defined as 'migration' host

    ############################################################################
#    def harmonize_cli(self, service, args):
#        """
#   - pull service target
#   - pull instance definitions from server
#     instance: type=persistent / container
#     if type=persistent:
#         login and run update on each instance (pull from s3)
#         in sequence, restart (indicate first), test for healthy and update
#     if type=container:
#         launch container
#         start new container, term old after it is healthy
#
#        """
#        self.NOTIFY("Harmonize")



#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END

# pylint: disable=missing-docstring

import os
import re
import traceback
import rfx
import rfx.client
from rfx.backend import EngineCli

################################################################################
class ControlCli(rfx.Base):

    rcs = None

    ############################################################
    # pylint: disable=super-init-not-called
    def __init__(self, base=None):
        if base:
            rfx.Base.__inherit__(self, base)
        if 'REFLEX_URL' in self.cfg.keys() and self.cfg['REFLEX_URL']:
            self.base_url = self.cfg['REFLEX_URL']
        self.rcs = rfx.client.Session(debug=self.debug, base=self)

    ############################################################
    def wizard_cli(self, argv, args): # pylint: disable=unused-argument
        for key in 'URL', 'APIKEY':
            if 'SET_REFLEX_' + key in os.environ.keys():
                self.cfg['REFLEX_' + key] = os.environ['SET_REFLEX_' + key]
            else:
                self.cfg['REFLEX_' + key] = input('REFLEX_' + key + ': ')
        self.cfg_save()

    ############################################################
    def set_cli(self, argv, args): # pylint: disable=unused-argument
        match = re.match(r'^([a-zA-Z_-]+)\s*=\s*(.*)$', " ".join(argv))
        if not match:
            self.NOTIFY("set var=value")
            return
        param = match.group(1)
        if match.group(2):
            value = match.group(2)
        else:
            value = input("Input Value: ")
        self.cfg[param] = value
        self.cfg_save()

    ############################################################
    def unset_cli(self, argv, args): # pylint: disable=unused-argument
        try:
            key = argv[0]
            del self.cfg[key]
            self.cfg_save()
        except KeyError:
            self.NOTIFY("parameter '" + key + "' not found.")
        except IndexError:
            self.NOTIFY("missing parameter name to unset.")

    ############################################################
    def list_cli(self, argv, args): # pylint: disable=unused-argument
        """dump the rfx config"""
        altcfg = self.cfg
        for line in altcfg:
            self.NOTIFY(line + "=" + altcfg[line])

    ############################################################
    def get_cli(self, argv, args): # pylint: disable=unused-argument
        """dump the rfx config"""
        self.timestamp = False
        if argv:
            self.OUTPUT(self.cfg[argv[0]])
        else:
            for line in self.cfg:
                self.OUTPUT("export " + line + "=" + self.cfg[line])

    ############################################################
    def apikey_cli(self, argv, args, cli):
        action = args.get('action')
        target = None
        if argv:
            target = argv[0]

        try:
            getattr(self, "apikey_cli__" + action)(target, cli)
        except rfx.CannotContinueError as err:
            err = str(err)
            if ": 401" in err:
                err = err + " Unauthorized"
            self.ABORT(err)

    ############################################################
    # pylint: disable=no-self-use,unused-argument
    def apikey_cli__delete(self, target, cli):
        self.rcs.delete("apikey", target)

    ############################################################
    def apikey_cli__list(self, target, cli):
        ecli = EngineCli(base=self)
        ecli.list_cli("apikey", {'--show': 'name'}, [])

    ############################################################
    def apikey_cli__create(self, target, cli):
        try:
            self.rcs.create("apikey", {"name":target})
            apikey = self.rcs.get('apikey', target)
            if not apikey:
                raise Exception("Unable to get object!")
            self.NOTIFY("new apikey:\n\n\t{}.{}"
                        .format(apikey.get('name', 'invalid'),
                                apikey.get('secrets', ['invalid'])[0]))
        except Exception: # pylint: disable=broad-except
            self.NOTIFY("Unable to properly create apikey!")
            if self.do_DEBUG():
                self.NOTIFY(traceback.format_exc(0))
                self.NOTIFY("(try --debug=* for more info)")
            else:
                self.NOTIFY(traceback.format_exc())



from setuptools import setup
setup(
  name = 'rfx',
  packages = ['rfx'],
  version = "1807.0003",
  description = 'Reflex - Container Config and Secret Management - core lib',
  author = 'Brandon Gillespie',
  author_email = 'bjg-pypi@cold.org',
  url = 'https://reflex.cold.org/',
  keywords = ['docker','config','secrets'],
  install_requires = [
    'urllib3',
    'dateparser',
    'ujson',
    'pynacl',
    'requests',
    'setproctitle',
    'dictlib',
    'onetimejwt',
    'timeinterval'
  ],
  classifiers = [],
)




#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END
"""
Launch Control
"""

import os
import traceback
import subprocess
import sys
import ujson as json
#import boto3
#from boto3.s3.transfer import S3Transfer
import rfx
from rfx import json4store
from rfx.backend import Engine

################################################################################
# pylint: disable=too-many-instance-attributes
class Action(rfx.Base):
    """
    Do actions, based on data stored in the immutable container
    """

    did = {} # to avoid recursion
    stack = []
    action_fqpath = '.'
    action_dir = ''
    action_file = ''
    action_dirs = ['.pkg/', '.reflex/', '.rfx', '.']
    config_files = ['config.json', 'actions.json']
    config = {}
    dbo = None
    extcfg = None
    s3cfg = None
    env = None

    ############################################################################
    def __init__(self, *args, **kwargs):
        super(Action, self).__init__(*args, **kwargs)
        if 'base' in kwargs: # sorry, we need the base info
            rfx.Base.__inherit__(self, kwargs['base'])

        # so any forked children don't inherit...
        if 'VIRTUAL_ENVIRON' in os.environ:
            virt = os.environ['VIRTUAL_ENVIRON'] + "/bin"
            newpath = []
            for elem in os.environ['PATH'].split(":"):
                if virt != elem:
                    newpath.append(elem)
            os.environ['PATH'] = ":".join(newpath)
            del os.environ['VIRTUAL_ENVIRON']

        self.get_terminal_width()
        self.env = os.environ.copy() # start with our os environ
        if 'PYTHONPATH' in self.env:
            del self.env['PYTHONPATH']

        self.default_environ('APP_RUN_BASE', os.getcwd())
        self.default_environ('APP_RUN_ROOT', self.env['APP_RUN_BASE'])
        self.default_environ('APP_CFG_BASE', os.getcwd())

        self.load_config()
        if 'test' not in self.debug:
            self.colorize = kwargs.get('colorize', True)
            try:
                self.get_my_nameip()
            except: # pylint: disable=bare-except
                pass

        if 'extcfg' in kwargs:
            self.extcfg = kwargs['extcfg']

    ############################################################################
    def default_environ(self, key, default):
        """set an action environ equal to os.environ(key) or the default value"""
        if key in os.environ:
            self.env[key] = os.environ[key]
        else:
            self.env[key] = default

    ############################################################################
    def load_config(self):
        """
        Load in the actions config file.

        Search path:

            ENV[APP_RUN_ROOT]/.pkg/actions.json
            (cwd)/.pkg/actions.json
        """

        paths = [os.getcwd()]
        env_root = self.env.get('APP_RUN_ROOT', '')
        if env_root:
            paths = [env_root + "/"] + paths

        def get_path():
            """Get path as a function so we can return once we find it"""

            for base_path in paths:
                for action_dir in self.action_dirs:
                    for conf_name in self.config_files:
                        base = base_path + "/" + action_dir + "/" + conf_name
                        if os.path.exists(base):
                            os.chdir(base_path)
                            path = base
                            self.action_fqpath = base_path + "/" + action_dir
                            self.action_dir = action_dir
                            self.action_file = conf_name
                            return path
            return None

        path = get_path()

        if not path:
            self.ABORT("No Reflex Actions configured, missing {}actions.json"
                       .format(self.action_dirs[0]))

        try:
            with open(path) as infile:
                self.config = json.load(infile)

        # pylint: disable=broad-except
        except Exception as err:
            self.DEBUG(traceback.format_exc())
            self.ABORT("Unable to load actions.json: {}".format(err))

    ############################################################################
    # pylint: disable=unused-argument
    def verify(self, target, export_meta=False):
        """
        Start at 'action' and walk the configs, verifying it is good.
        """
        stack = " -> ".join(self.stack)
        if target in self.did:
            self.ABORT("Recursive action calls back in on itself! (" + stack + ")")

        actions = self.config['actions']

        if target not in actions:
            self.ABORT("Cannot find target action '" + target + "' (" + stack + ")")

        if 'type' not in actions[target]:
            self.ABORT("Cannot find action type for '" + target + "' (" + stack + ")")

        if actions[target]['type'] == 'run':
            if not os.path.exists(self.action_fqpath + "/" + actions[target]['target']):
                self.ABORT("Cannot find action script {} for '{}' ({})"
                           .format(self.action_fqpath + "/" + actions[target]['target'],
                                   target, stack))

        self.did[target] = True
        self.stack.append(target)
        self.DEBUG("Verified " + target)

        if 'onSuccess' in actions[target]:
            self.verify(actions[target]['onSuccess'])

    ############################################################################
    def _do(self, target, opts=None):
        """
        do an action (recursive)

        Assert: self.verify should already have been run, so we can assume no
        recursion and our data structures are good.
        """

        action = self.config['actions'][target]

        label = action.get('title', target)

        self.banner("Action '" + label + "'")

        env = self._do_setenv(action)

        # could also do exec
        if action['type'] == 'script':
            if 'target' in action:
                path = self.action_fqpath + '/' + action['target']
            else:
                path = self.action_fqpath + '/' + target
            cmd = [path]
            if opts:
                cmd += opts
            return self._do__cmd(target, action, cmd, env=env)
        elif action['type'] == 'system':
            # backwards compatible, deprecated use of "cmd" vs "cmd"
            cmd = action.get("exec", None)
            if not cmd:
                cmd = action.get("cmd")
            if opts:
                cmd += opts
            return self._do__cmd(target, action, cmd, env=env)
        elif action['type'] == 'exec':
            cmd = action.get("cmd")
            if opts:
                cmd += opts
            return self._do__cmd(target, action, cmd, env=env, doexec=True)
        else:
            try:
                func = getattr(self, '_do__' + action['type'].replace('-', '_'))
            except AttributeError:
                self.NOTIFY("Ignoring unrecognized action: " + action['type'])
                return False
            return func(target, action)

    ############################################################################
    def _do__docker_service(self, target, action):
        stype = action.get("system", "systemctl")
        sysaction = action.get("action", "start")
        if stype == "systemctl":
            cmd = ["sudo", "/bin/systemctl", sysaction, "docker"]
        else:
            self.NOTIFY("Ignoring unrecognized system type (systemctl)")
            return False

        return self._do__cmd(target, action, cmd, echo=True)

    def _do__docker_build(self, target, action):
        """
        Store a package somewhere. Currently only supports s3.

        Run after roll_package.  Assumes LAST_PKG_FILENAME is in environ setting.
        """

        args = []
        if self.env.get("NO_DOCKER_CACHE"):
            args = ["--no-cache"]
        version = self.env.get('PKG_VERSION')
        if version:
            version = ":" + version
        image = action.get("image")
        if not image:
            self.NOTIFY("No image name defined!")
            return False
        image = self.sed_env(image, {}, '', env=self.env)
        tags = list()
        for tag in action.get('tags', []):
            tags.append("-t")
            tags.append(image + ":" + tag)

        dockerfile = action.get("dockerfile")
        if not dockerfile:
            if os.path.exists("Dockerfile"):
                dockerfile = "Dockerfile"
            elif os.path.exists(self.action_dir + "/Dockerfile"):
                dockerfile = self.action_dir + "/Dockerfile"

        if not os.path.exists(dockerfile):
            self.ABORT("Unable to fid a Dockerfile")

        return self._do__cmd(target, action,
                             ["docker", "build"] + args +
                             ["-t", image + version] +
                             tags +
                             ["-f", dockerfile, "."], echo=True)

    def _do__docker_push(self, target, action):
        """
        Store a package somewhere. Currently only supports s3.

        Run after roll_package.  Assumes LAST_PKG_FILENAME is in environ setting.
        """

        image = action.get("image")
        if not image:
            self.NOTIFY("No image name defined!")
            return False

        cmd = ["docker", "push", image]
        return self._do__cmd(target, action, cmd, echo=True)

    ############################################################################
    def _do__store_package_s3(self, target, action):
        """
        Store a package somewhere. Currently only supports s3.

        Run after roll_package.  Assumes LAST_PKG_FILENAME is in environ setting.
        """

        fname = self.env.get('LAST_PKG_FILENAME', 'unknown-file')
        try:
            repo = self.config['config']['package-repo']
        except AttributeError:
            self.ABORT("Cannot find 'package-repo' in actions.json->config")

        self.NOTIFY("Storing " + fname + " to " + repo)
        return self._do__cmd(target, action, [
            "s3cmd", "--no-mime-magic", "--no-progress", "--quiet", "put", fname, repo + '/' + fname
        ], env={})

        # more work w/S3 to polish this
#        if repo[:5] == 's3://':
#            repo = repo[5:]
#        s3 = S3Transfer(boto3.client('s3'))
#        response = s3.upload_file(fname, repo, fname)
#        print(response)

    ############################################################################
    def _do__roll_package(self, target, action):
        """
        Based on provided include/exclude files, create package.
        """

        try:
            name = "deploy_stage." + \
                   self.env.get("REFLEX_PRODUCT", 'NO_REFLEX_PRODUCT') + "_" + \
                   self.env.get("REFLEX_MODULE", 'NO_REFLEX_MODULE')
            include = self.action_fqpath + action['include']
        except KeyError as err:
            self.ABORT("Unable to roll package '{}', missing attribute: {}".format(target, err))

        try:
            pkgfile = name + '.' + self.env.get('PKG_VERSION', 'NO_PKG_VERSION') + '.tgz'
        except KeyError as err:
            self.ABORT("Unable to roll package '{}', missing setenv: {}".format(target, err))

        cmd = self._sh("rm -f " + name + "*.tgz")
        if cmd['code'] > 0:
            self.OUTPUT(cmd['out'])

        if not os.path.exists(include):
            self.ABORT("Cannot find include file: " + include)

        exc = ["tar", "-cvzf", pkgfile]

        if action.get('chdir', False):
            if os.path.exists(action['chdir']):
                exc += ["--directory=" + action['chdir']]

        # order matters -- must come after chdir
        exc += ["--files-from=" + include]

        if action.get('gitignore', False):
            if os.path.exists(".gitignore"):
                exc += ["--exclude-from=" + os.getcwd() + "/.gitignore"]

        if action.get('exclude', None):
            exc += ["--exclude-from=" + self.action_fqpath + action['exclude']]

        cmd = self._run(exc)
        if cmd['code'] > 0:
            self.ABORT("Cannot continue:\n" + cmd['out'])

        cmd = self._run(['gzip', '-t', pkgfile])
        if cmd['code'] > 0:
            self.ABORT("Unable to verify package:\n" + cmd['out'])

        # change the default environ
        self.env['LAST_PKG_FILENAME'] = pkgfile

        return self._do_success(target, action)

    ############################################################################
    def _sh(self, cmd, spliterr=False):
        """
        Wrap Popen with common usable features.  Similar to 'check_output' but
        the way I want it.

        >>> self = Action()
        >>> import json
        >>> json.dumps(self._sh("echo ohhi"), sort_keys=True)
        '{"code": 0, "err": "", "out": "ohhi\\\\n"}'
        """
        return self._run(cmd, spliterr=spliterr, shell=True)

    ############################################################################
    def _run(self, *args, **kwargs): #spliterr=False):
        """
        Wrap Popen with common usable features.  Similar to 'check_output' but
        the way I want it.

        Output from python 2.7: {'code': 0, 'err': '', 'out': 'hi\\n'}

        >>> self = Action()
        >>> import json
        >>> json.dumps(self._run(["echo", "hi"]), sort_keys=True)
        '{"code": 0, "err": "", "out": "hi\\\\n"}'
        """
        spliterr = kwargs.get('spliterr', False)
        if spliterr:
            stderr = subprocess.PIPE
        else:
            stderr = subprocess.STDOUT
        shell = False
        if 'shell' in kwargs:
            shell = kwargs['shell']

        exc = list(args)[0]
        self.DEBUG("run: {}".format(exc))
        self.notifyfd.flush()
        self.outputfd.flush()
        sub = subprocess.Popen(exc, shell=shell, stdout=subprocess.PIPE,
                               stderr=stderr, env=self.env)
        output, outerr = sub.communicate()
        if isinstance(output, bytes): # grr python 2/3
            output = output.decode()
        if isinstance(outerr, bytes): # grr python 2/3
            outerr = outerr.decode()

        return {'code': sub.returncode, 'out': output, 'err': outerr or ''}

    ############################################################################
    def _do__group(self, ignore, action):
        """
        Run a group of actions in sequence.
        """

        for target in action.get('actions', []):
            self.do(target)

    ################################################################################
    # evaluate to fully qualified exe, incase env={} and default PATH
    def _executable(self, exe):
        """logic wrapper to simplify finding execuable"""
        if exe[0][:1] != "/" and exe[0][:2] != "./":
            path = self.env.get('PATH', None)
            if path is None:
                return False
            for path in os.environ.get('PATH', '').split(os.pathsep):
                fqpath = os.path.join(path, exe[0])
                if os.access(fqpath, os.X_OK):
                    exe[0] = fqpath
                    return exe
        elif os.access(exe[0], os.X_OK):
            return exe
        return False

    ############################################################################
    # pylint: disable=too-many-arguments,dangerous-default-value,too-many-branches,inconsistent-return-statements
    def _do__cmd(self, target, action, exc, env=dict(), echo=False, doexec=False):
        """
        Execute a sub process as part of a action, and handle the subsequent step

        if env=False, do not use the reflex environ settings, just the default os
        """

        if not env:
            env = self.env

        # convert any ${vars} in exe list
        env_exc = []
        for elem in exc:
            env_exc.append(self.sed_env(elem, {}, '', env=env))

        # pull first arg from env{PATH}
        fqexc = self._executable(env_exc)
        if not fqexc:
            raise ValueError("Cannot execute: {}".format(exc[0]))

        extcfg = ''
        if action.get('config', "n/a") == "stdin" and self.extcfg:
            if isinstance(self.extcfg, str):
                extcfg = json4store(self.extcfg)
            else: # the App object exports a dictionary
                extcfg = json4store(self.extcfg.export())

        if echo:
            formatted = []
            for arg in fqexc:
                if ' ' in arg:
                    formatted.append(json.dumps(arg))
                else:
                    formatted.append(arg)
            self.NOTIFY("Execute:\n\n\t" + " ".join(formatted) + "\n")

        self.DEBUG("command: " + str(fqexc))
        self.notifyfd.flush()
        self.outputfd.flush()
        if doexec:
            # pylint: disable=inconsistent-return-statements
            # this is the end
            os.execv(fqexc[0], fqexc)
        else:
            proc = subprocess.Popen(fqexc, stdin=subprocess.PIPE, env=env)
            if extcfg:
                proc.stdin.write(extcfg.encode())
            proc.stdin.close()
            proc.wait()
            if proc.returncode > 0:
                return self._do_failure(target, action, proc.returncode)
            return self._do_success(target, action)

    ############################################################################
    # pylint: disable=unused-argument
    def _do_failure(self, target, action, exitstatus):
        """
        Handle a failure
        """
        self.ABORT("Unable to continue after failure on '" + target + "'")

    ############################################################################
    # pylint: disable=unused-argument,inconsistent-return-statements
    def _do_success(self, target, action):
        """
        Handle a success
        """
        if 'onSuccess' in action:
            return self._do(action['onSuccess'])

    ############################################################################
    # pylint: disable=unused-argument
    def _do_setenv(self, cfg, output=False, export_meta=False):
        """
        Process a 'setenv' block, substituting $VAL from environment.

        if output=True or debug, print export output
        """
        if export_meta:
            for key in 'APIKEY', 'URL':
                key = 'REFLEX_' + key
                if key not in self.env:
                    self.env[key] = self.cfg[key]

        if not 'setenv' in cfg:
            return

        # use sys.stdout directly so the i/o order is correct
        # should consider changing to macro_expand (or making the % vs $ be variable
        # on sed_env.
        mdict = self.union_dict(cfg['setenv'], cfg.get('config', {}))
        for key in sorted(cfg['setenv']):
            self.env[key] = self.sed_env(str(cfg['setenv'][key]), mdict, key, env=self.env)
            if output:
                self.OUTPUT("export {}=\"{}\"".format(key, self.env[key]))
            elif self.do_DEBUG(module='test'):
                sys.stdout.write("export {}=\"{}\"\n".format(key, self.env[key]))
        sys.stdout.flush()
        return

    ############################################################################
    def banner(self, msg):
        """
        Print a distinguishable banner
        """

        if self.logfmt == 'txt':
            if self.do_DEBUG(module="test"):
                width = 80
            else:
                width = self.term_width
            if self.timestamp:
                width -= len(self.TIMESTAMP())
            banner_fmt = "{:=>" + str(width) + "s}"
            self.NOTIFY(banner_fmt.format(''), color='blue')
            padlen = width - (len(msg) + 6) # can overrun
            if padlen < 0:
                padlen = 0
            self.NOTIFY("==[ {} ]{:=>{len}s}".format(msg, '', len=padlen), color='blue')
            self.NOTIFY(banner_fmt.format(''), color='blue')
        else:
            self.NOTIFY(msg)

    ############################################################################
    # pylint: disable=invalid-name,inconsistent-return-statements
    def do(self, target, opts=None, export_meta=False):
        """
        do an action
        """

        # this is true if we have verified it
        if target not in self.did:
            self.verify(target, export_meta=export_meta)

        rfx.unbuffer(self.notifyfd)
        rfx.unbuffer(self.outputfd)

        # reset after ^^
        self.did = {}
        self.stack = []

        msg = "Starting Action '" + target + "'"
        if self.my_ip:
            msg += " on " + self.my_ip
        self.banner(msg)

        self.env['REFLEX_ACTION_BASE'] = self.action_fqpath

        # build environment, reset default to this
        self._do_setenv(self.config, export_meta=export_meta)

        try:
            return self._do(target, opts=opts)
        except KeyboardInterrupt:
            return

    ############################################################################
    def db_connect(self):
        """
        Open a connection to the back-end reflex-engine
        """
        if self.dbo:
            return
        self.dbo = Engine(base=self)



#$#HEADER-START
# vim:set expandtab ts=4 sw=4 ai ft=python:
#
#     Reflex Configuration Event Engine
#
#     Copyright (C) 2016 Brandon Gillespie
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#$#HEADER-END
# pylint: disable=missing-docstring

"""
Configuration Object Fu
"""

import re
try:
    import StringIO as strio # pylint: disable=import-error
except: # pylint: disable=bare-except
    import io as strio # pylint: disable=reimported,ungrouped-imports
import os
import base64
import copy
import rfx
from rfx import json4store
import dictlib

############################################################
def deep_get(obj, key):
    """
    >>> deep_get({"a":{"b":{"c":1}}}, "a.b.c")
    1
    """
    array = key.split(".")
    return _deep_get(obj, *array)

def _deep_get(obj, *key):
    """
    Recursively lookup an item in a nested dictionary,
    using an array of indexes

    >>> _deep_get({"a":{"b":{"c":1}}}, "a", "b", "c")
    1
    """
    if len(key) == 1:
        return obj[key[0]]
    return _deep_get(obj[key[0]], *key[1:])

################################################################################
class VerboseBase(rfx.Base):
    verbose = True

    def __init__(self, *args, **kwargs):
        if not kwargs.get('verbose'):
            self.verbose = False
        super(VerboseBase, self).__init__(*args, **kwargs)

    # pylint: disable=invalid-name,inconsistent-return-statements
    def NOTIFY(self, *msg, **kwargs):
        if self.verbose:
            return super(VerboseBase, self).NOTIFY(*msg, **kwargs)

    # pylint: disable=invalid-name
    def DEPRECATE(self, msg):
        self.NOTIFY("DEPRECATED: " + msg)

################################################################################
class ConfigProcessor(VerboseBase):
    """
    Process configuration objects.  Keeps state and does recursion.

    .flatten() will bring everything into one object, with variables substituted
    .commit() will take the object and put it out to disk

    Notes on data elements:

    - extends:
        * processed by .flatten()
        * goes up a heirarchy, recursively
        * preference given to values lower in tree (child overrides parent)
        * ignored on exported target
    - imports:
        * processed by .flatten()
        * merges a single object into current
        * preference given to values on imported object, not current
        * imports does not merge 'extends' nor 'imports' values
        * can be used on exported targets
    - exports:
        * processed by .commit(), which also re-triggers .flatten() on exported objects
        * target objects are merged with original object
        * preference given to target object values
        * extends is not processed on target objects (no hierarchy processed)
        * imports is processed on target objects
        * exported objects should be type=file, otherwise nothing happens
    """

    cfgdir = ''
    did = None
    rx_var = re.compile(r"%\{([a-zA-Z0-9_.-]+)\}") # macro_expand
    rx_env = re.compile(r"\$(\{([a-zA-Z0-9_]+)\}|([a-zA-Z0-9_]+))") # environ_var
    expanded_vars = None
    rcs = None
    peers = None

    ############################################################################
    def __init__(self, verbose=True, base=None, rcs=None, peers=None):
        super(ConfigProcessor, self).__init__(verbose=verbose)
        if not base: # sorry, we need rfx.Base
            raise ValueError("Missing Reflex Base (base=)")
        if not rcs: # and rcs
            raise ValueError("Missing Reflex Engine (rcs=)")
        rfx.Base.__inherit__(self, base)
        self.rcs = rcs # allows for overriding during testing
        self.verbose = verbose
        self.did = dictlib.Obj(exp=dict(), imp=dict(), ext=dict())
        if peers:
            self.peers = peers
        else:
            self.peers = dict()


    ############################################################################
    # pylint: disable=invalid-name,inconsistent-return-statements
    def NOTIFY(self, *msg, **kwargs):
        if self.verbose:
            return super(ConfigProcessor, self).NOTIFY(*msg, **kwargs)

    ############################################################
    def macro_expand_dict(self, dictionary):
        """
        For every entry in dictionary, process macro's on all
        values, using the dict as the lookup table.

        >>> base = rfx.Base()
        >>> test = ConfigProcessor(base=base, rcs=base)
        >>> test.macro_expand_dict({"a":"Hello",
        ...                         "b":"Not %{a}",
        ...                         "c":"c=%{c}",
        ...                         "d":"more %{b}"})
        {'a': 'Hello', 'c': 'c=', 'b': 'Not Hello', 'd': 'more Not Hello'}
        """
        for key in dictionary:
            dictionary[key] = self.macro_expand(dictionary[key], dictionary, source=key)
        return dictionary

    ############################################################

    def macro_expand(self, value, dictionary, source=None):
        """
        Expand our macro structure, referencing dictionary for
        replacement values.  Value may contain multiple macros.

        >>> base = rfx.Base()
        >>> test = ConfigProcessor(base=base, rcs=base)
        >>> test.macro_expand("hi %{VAR}", {"VAR":"there"})
        'hi there'

        >>> test.macro_expand("hi %{VAR}", {"VAR":"more %{T}", "T":"there"})
        'hi more there'

        >>> test.macro_expand("hi %{NOT_A_VAR}", {})
        'hi '

        >>> os.environ['NOT_A_VAR'] = 'from environ'
        >>> test.macro_expand("hi %{NOT_A_VAR}", {"NOT_A_VAR": "from dict"})
        'hi from environ'
        """

        def do_match(match):
            """Internal function for processing a matched variable"""
            match_key = match.group(1)
            if match_key in os.environ:
                return os.environ[match_key]
            dict_match = dictlib.dig_get(dictionary, match_key, None)
            if match_key != source and dict_match != None:
                return str(dict_match)
            self.NOTIFY("Unable to find expansion match for key '" + match_key + "'")
            return None

        # loop to recurse properly (handling macros in macros)
        if isinstance(value, bytes):
            value = value.decode()
        if not isinstance(value, str):
            value = str(value)
        # first do our configs
        while self.rx_var.search(value):
            value = self.rx_var.sub(do_match, value)
        return value

    ############################################################################
    def flatten(self, name, exported=False):
        conf = Config(name, base=self, rcs=self.rcs, verbose=self.verbose).conf
        conf = self._flatten(conf, name, exported=exported, toplevel=True)
        return self._process(conf)

    ############################################################################
    def _process(self, conf):
        subname = (conf.name + ".").split(".")[0]
        self.NOTIFY("Processing config object {0}".format(subname))
        # process the variables
        self.expanded_vars = allvars = dict()
        procvars = set(['sensitive.parameters'] + conf.procvars)

        # pass1, merge all vars into one monster dictionary
        for key in procvars:
            dictlib.union(allvars, deep_get(conf, key))

        # pass2 expand all variables
        self.macro_expand_dict(allvars)

        # pass3 put values back in their place
        for target in procvars:
            tobj = deep_get(conf, target)
            if not isinstance(tobj, dict):
                self.ABORT("Target {} is not a dictionary".format(dict))
            for key in tobj:
                tobj[key] = allvars[key]

        # pass4, do setenv last, it does not merge into allvars view
        for key in conf.setenv:
            conf.setenv[key] = self.macro_expand(conf.setenv[key], allvars)
            #conf.setenv[key] = self.sed_env(conf.setenv[key], allvars, key)

        if conf.content.get('dest'):
            conf.content.dest = self.macro_expand(conf.content.dest, allvars)
        if conf.get('file'):
            self.DEPRECATE("conf.file should be conf.content.dest")
            conf.file = self.macro_expand(conf.file, allvars)

        if conf.content.get('source'):
            conf.content.source = self.macro_expand(conf.content.source, allvars)

        return conf

    ############################################################################
    def _flatten(self, conf, name, exported=False, toplevel=False):
        """Merge in all relatives"""
        if name in self.did.ext:
            return conf
        self.did.ext[name] = True

        new = Config(name, base=self, rcs=self.rcs, verbose=self.verbose).load()

        def dmerge_outer(conf, new, key):
            if key in new and new[key]:
                conf[key] = dictlib.union(conf[key], new[key])

        def dmerge_inner(conf, new, key):
            if key in new and new[key]:
                conf[key] = dictlib.union(new[key], conf[key])

        def lmerge(conf, new, key):
            if key in new and new[key]:
                # sets are not ordered, so do it by hand; set-add old array to new
                for item in conf[key]:
                    if not item in new[key]:
                        new[key].append(item)
                conf[key] = new[key]

        lmerge(conf, new, 'imports')

        # single level inheritance
        if conf.imports:
            for iname in conf.imports:
                if iname in self.did.imp:
                    continue
                self.did.imp[iname] = True
                iconf = Config(iname, base=self, rcs=self.rcs, verbose=self.verbose).load()

                # imports ignores `extends` and `imports`
                dmerge_outer(conf, iconf, 'sensitive')
                dmerge_outer(conf, iconf, 'setenv')
                dmerge_outer(conf, iconf, 'content')
                if not exported:
                    lmerge(conf, iconf, 'exports')

        # and merge into self
        lmerge(conf, new, 'extends')
        dmerge_inner(conf, new, 'sensitive')
        dmerge_inner(conf, new, 'setenv')
        if not exported:
            lmerge(conf, new, 'exports')

        # inheriting content is difficult
        # if we are top level or exporting, give that precedence
        if exported or toplevel:
            dmerge_outer(conf, new, 'content')

        # otherwise, check per value
        elif new.content and not conf.content:
            dmerge_inner(conf, new, 'content')
#            def content_var(name):
#                if new.content.get(name) and not conf.content.get(name):
#                    conf.content[name] = new.content.get(name)
#            content_var('dest')
#            content_var('source')
#            content_var('type')
#            content_var('varsub')
#            content_var('encoding')
#            content_var('ref')

        if new.file and not conf.file:
            conf.file = new.file

        lmerge(conf, new, 'procvars')

        if not exported:
            for parent in new.extends:
                self._flatten(conf, parent)

        # set type last
        conf.type = new.type
        return conf

    ############################################################################
    def commit(self, conf, dest=None, exported=False):
        """
        Update environment and store files.
        Split into two steps so we can discretely call setenv without exporting files.
        """
        self.prep_setenv(conf, dest=dest, exported=exported)
        self.export_files(conf, dest=dest, exported=exported)

    ############################################################################
    # pylint: disable=unused-argument
    def prep_setenv(self, conf, dest=None, exported=False):
        """export any files"""
        if dest: # override default
            self.cfgdir = dest
        elif 'APP_CFG_BASE' in os.environ:
            self.cfgdir = os.environ['APP_CFG_BASE']
        else:
            raise ValueError("APP_CFG_BASE not defined")

        def envset(key, value):
            if key in conf.setenv and conf.setenv[key] != value:
                self.NOTIFY("Overriding existing setenv: " + key)
            conf.setenv[key] = value

        def envset_from_os(key, default):
            envset(key, os.environ.get(key, default))

        envset('APP_CFG_BASE', self.cfgdir)
        envset_from_os('APP_RUN_BASE', '')
        envset_from_os('APP_PIPELINE', '')
        envset_from_os('APP_SERVICE', '')
        envset_from_os('TMPDIR', '.')

        for key, value in conf.setenv.items():
            os.environ[key] = str(value)

    ############################################################################
    # pylint: disable=unused-argument
    def export_files(self, conf, dest=None, exported=False):
        """
        Export Files.
        """
        # export any objects
        if not exported:
            for name in conf.exports:
                if name in self.did.exp:
                    continue
                self.did.exp[name] = True
                exp = copy.deepcopy(conf)
                exp.name = name
                exp = self._flatten(exp, name, exported=True)
                exp = self._process(exp)
                self.commit(exp, dest=self.cfgdir, exported=True)

        # and process it
        if conf.type == 'file':
            self._commit_file(conf)
        else:
            self.DEBUG("Nothing to commit on {name}.type='{type}'".format(**conf))

    ############################################################################
    # pylint: disable=no-self-use
    def _decode_none(self, in_f):
        """Return contents of a file, with no decoding"""
        return in_f.read()

    ############################################################################
    # pylint: disable=no-self-use
    def _decode_base64(self, in_f):
        """Return contents of a file, with base64 decoding"""
        return base64.b64decode(in_f.read())

    ############################################################################
    def _commit_file(self, conf):
        """Process a conf, as a file type (store output)"""
        if not conf.content:
            raise ValueError("Object '{name}' missing 'content'".format(**conf))

        content = conf.content

        if content.get('dest'):
            fname = content.dest
        elif conf.get('file'):
            fname = conf.file
        else:
            raise ValueError("Cannot find output filename ({name}.content.dest)"
                             .format(**conf))

        # for security all files are relative to APP_CFG_BASE, no
        # .. paths allowed.
        if ".." in fname:
            raise ValueError("File name cannot contain ..")

        fname = self.cfgdir + "/" + fname

        # using ref allows us to store the data in "sensitive"
        # (data at rest) area:
        if content.get('ref'):
            buf = self._get_content_ref(conf)
        elif content.get('source'):
            buf = self._get_content_src(conf)
        else:
            raise ValueError("Cannot find file content for object " + conf.name)

        if content.get('varsub'):
            buf = self.macro_expand(buf, self.expanded_vars)

        if isinstance(buf, str): # writing everything binary
            buf = buf.encode()

        subname = (conf.name + ".").split(".")[0]
        self.NOTIFY("CONFIG {0} into {1}".format(subname, fname))
        with open(fname, 'wb') as ofile:
            ofile.write(buf)

    def _get_content_ref(self, conf):
        ref = conf.content.ref
        if isinstance(ref, str):
            data = deep_get(conf, ref)
        elif isinstance(ref, list):
            self.DEPRECATE("conf.content.ref should be a string in x.z.y notation.")
            data = _deep_get(conf, *ref)
        else:
            raise ValueError("Unrecognized data type for conf.content.ref ({})"
                             .format(type(conf.content.ref)))

        if conf.content.get('type', 'text/plain') == 'application/json':
            if isinstance(data, dictlib.Obj):
                data = data.__export__()
            data = json4store(data)

        # pylint: disable=line-too-long
        return getattr(self, "_decode_" + conf.content.get('encoding', 'none'))(strio.StringIO(data))

    def _get_content_src(self, conf):
        infile = self.cfgdir + "/" + conf.content.source
        if not os.path.exists(infile):
            raise ValueError("Unable to process config file: " + infile)
        with open(infile, 'rt') as infile:
            content_f = strio.StringIO(infile.read())

        return getattr(self, "_decode_" + conf.content.get('encoding', 'none'))(content_f)

    def prune(self, conf):
        """
        Cleanup the config, removing extraneous bits.

        Run /after/ commit.
        """

        def ifdel(key):
            """shortcut"""
            if key in conf:
                del conf[key]

        ifdel('file')
        ifdel('procvars')
        ifdel('extends')
        ifdel('imports')
        ifdel('exports')
        ifdel('type')
        ifdel('macroExpand')

        return conf

################################################################################
class Config(VerboseBase):
    """
    Manage configurations

    .get/put/etc -- just the named object
    .dump() -- return a json form << replaces .export() >>
    """

    conf = None
    rcs = None

    ############################################################################
    def __init__(self, name, base=None, rcs=None, verbose=True):
        super(Config, self).__init__(verbose=verbose)
        if not base: # we need rfx.Base
            raise ValueError("Missing Reflex Base (base=)")
        rfx.Base.__inherit__(self, base)
        if not rcs: # and a rfx.Engine() object
            raise ValueError("Missing Reflex Engine (rcs=)")

        self.conf = dictlib.Obj(
            name=name,
            sensitive=dictlib.Obj(
                parameters=dict(
                )
            ),
            exports=list(),
            imports=list(),
            extends=list(),
            setenv=dict(),
            content=dict(),
            procvars=list(),
            file='',
            type='parameter'
        )

        self.rcs = rcs
        self.peers = base.peers # should come in from ConfigProcessor
        if self.peers:
            self.conf['peers'] = self.peers
            for iplabel in ('0', '1'):
                peers = self.peers['ip' + iplabel]

                self.conf.setenv['LAUNCH_PEERS' + iplabel] = \
                    ",".join([key + "@" + value for key, value in peers.items()])
                self.conf.setenv['LAUNCH_PEER' + iplabel + '_NAMES'] = ",".join(list(peers.keys()))
                self.conf.setenv['LAUNCH_PEER' + iplabel + '_IPS'] = ",".join(list(peers.values()))

    def load(self):
        obj = self.rcs.get('config', self.conf.name)
        if obj:
            dictlib.union(self.conf, obj)

        if 'export' in self.conf:
            self.DEPRECATE("conf.export should be conf.exports")
            self.conf.exports = self.conf.export

        if 'macroExpansion' in self.conf:
            self.conf.content['varsub'] = self.conf.macroExpansion
            self.DEPRECATE("conf.macroExpansion should be conf.content.varsub")

        return self.conf

    ############################################################################
    def __repr__(self):
        """Return a text representation of the object"""
        return json4store(self.conf)

